from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Union, Optional
from uuid import uuid4
import time

from pydantic import (
    BaseModel, 
    ConfigDict, 
    field_serializer,
    Field,
    model_validator
    )

from autogen_core import CancellationToken, ComponentBase


class TaskStatus(str, Enum):
    """The status of a task in the task management system."""

    queued = "queued"
    in_progress = "in_progress"
    completed = "completed"
    paused = "paused"
    error = "error"

class TaskType(str, Enum):
    """The type of a task in the task management system."""

    add = "add"
    insert = "insert"
    delete = "delete"
    update = "update"
    select = "select"


class Task(BaseModel):
    """A task in the task management system."""

    content: str
    """The content of the task."""

    source: str
    """The source of the task."""

    id: str = Field(default_factory=lambda: str(uuid4()))
    """The unique identifier of the task."""

    qid: Optional[str] = None
    """queue id in task management system."""

    execution_count: int = 0
    """execution count of the task."""

    discussion_round: int = 0
    """discussion count of the task."""

    task_type: TaskType|str = TaskType.add
    """The type of the task."""
    
    status: TaskStatus|str = TaskStatus.queued
    """The status of the task."""

    created_at: float = Field(default_factory=lambda: time.time())
    """The time when the task was created."""

    updated_at: Optional[float] = Field(default=None)
    """The time when the task was last updated."""

    parent_task_id: Optional[str] = None
    """The parent task id of the task."""

    child_task_ids: List[str] = []
    """The child task ids of the task."""

    child_tasks: List[Dict[str, Any]] = []

    metadata: Dict[str, Any] | None = None
    """Metadata associated with the memory item."""

    completed_at: Optional[float] = Field(default=None)
    """The time when the task was completed."""

    executor: Optional[str] = None
    """The executor of the task."""

    solution: Optional[str] = None
    """The solution of the task."""


    @model_validator(mode='after')
    def set_timestamps(self):
        if self.updated_at is None:
            self.updated_at = self.created_at
        return self

    def update(self, **kwargs):
        """Update task attributes and automatically update the updated_at timestamp."""
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
        # Update the updated_at timestamp
        self.updated_at = time.time()
        return self

    @field_serializer("status")
    def status_serializer(self, status: TaskStatus | str) -> str:
        if isinstance(status, TaskStatus):
            return status.value
        else:
            return status

class BaseTaskSystem(ABC, ComponentBase[BaseModel]):

    """The base class for task management systems in group chat teams.

    Basic methods for task management:
    - create_task: create a new task with the given task_id and task_data
    - get_task: get the task with the given task_id
    - update_task: update the task with the given task_id and task_data
    - delete_task: delete the task with the given task_id
    - list_tasks: list all tasks in the system
    """

    component_type = "task_system"

    @abstractmethod
    async def create_task(
        self, 
        task_content: str, 
        metadata: Dict[str, Any] | None = None, 
        cancellation_token: CancellationToken | None = None,
        **kwargs: Any,
        ) -> Task:
        """Create a new task with the given task_content and metadata."""
        ...
    
    @abstractmethod
    async def get_task(
        self, 
        task_id: str, 
        cancellation_token: CancellationToken | None = None,
        **kwargs: Any,
        ) -> Task:
        """Get the task with the given task_id."""
        ...
    
    @abstractmethod
    async def update_task(
        self, 
        task_id: str, 
        task_content: str | None = None, 
        metadata: Dict[str, Any] | None = None, 
        status: TaskStatus | str | None = None, 
        cancellation_token: CancellationToken | None = None,
        **kwargs: Any,
        ) -> Task:
        """Update the task with the given task_id and task_data."""
        ...
    
    @abstractmethod
    async def delete_task(
        self, 
        task_id: str, 
        cancellation_token: CancellationToken | None = None,
        **kwargs: Any,
        ) -> None:
        """Delete the task with the given task_id."""
        ...
    
    @abstractmethod
    async def list_tasks(
        self, 
        status: TaskStatus | str | None = None, 
        cancellation_token: CancellationToken | None = None,
        **kwargs: Any,
        ) -> List[Task]:
        """List all tasks in the system."""
        ...
    
    @abstractmethod
    async def reset(self, **kwargs: Any) -> None:
        """Reset the task management system."""
        ...



from pydantic import BaseModel
from abc import ABC, abstractmethod
from dataclasses import dataclass
import asyncio
from typing import Any, Dict, List, Optional, Union, Callable
from autogen_core import (
    ComponentModel,
    ComponentBase,
    Component,
    CancellationToken,
    AgentRuntime,
    # MessageContext, 
    # event, 
    # rpc,
)
from autogen_agentchat.base import (
    ChatAgent, 
    )
from autogen_agentchat.teams._group_chat._events import (
    GroupChatAgentResponse,
    GroupChatError,
    GroupChatMessage,
    GroupChatPause,
    GroupChatRequestPublish,
    GroupChatReset,
    GroupChatResume,
    GroupChatStart,
    GroupChatTermination,
    SerializableException,
    )
GroupChatEvent = (GroupChatStart|
                  GroupChatAgentResponse|
                  GroupChatError|
                  GroupChatMessage|
                  GroupChatPause|
                  GroupChatRequestPublish|
                  GroupChatReset|
                  GroupChatResume|
                  GroupChatTermination|
                  SerializableException)

from drsai.modules.groupchat.base_group_chat_runner import BaseGroupChatRunner

def handle_group_chat_event(func: Callable[[GroupChatEvent], Any]) -> None:
    ...


class BaseRunTimeDeamon(ABC, ComponentBase[BaseModel]):
    """ 
    A runtime deamon to manage GroupChatRunner running tasks.
    """
    
    component_type = "run_time_deamon"

    @abstractmethod
    async def send_message(
        self,
        message: Any,
        sender_id: str,
        recipient_id: str,
        *,
        cancellation_token: CancellationToken | None = None,
        ) -> Any:
        """
        Send a message to the group chat.
        """
        ...
    
    async def publish_message(
        self,
        message: Any,
        sender_id: str,
        recipient_id: str,
        *,
        cancellation_token: CancellationToken | None = None,
        ) -> Any:
        """
        Publish a message to the group chat.
        """
        ...

    @abstractmethod
    async def get_GroupChatRunner_instance(
        self,
        group_chat_id: str,
        group_chat_runner_factory: Callable[[], BaseGroupChatRunner]|None = None,
        agent_factories: Dict[str, Callable[[], ChatAgent]]|None = None,
    ) -> Any:
        """
        Get a GroupChatRunner instance for the given group chat.
        """
        ...


# class RunContext:
#     def __init__(self, runtime: SingleThreadedAgentRuntime) -> None:
#         self._runtime = runtime
#         self._run_task = asyncio.create_task(self._run())
#         self._stopped = asyncio.Event()

#     async def _run(self) -> None:
#         while True:
#             if self._stopped.is_set():
#                 return

#             await self._runtime._process_next()  # type: ignore

#     async def stop(self) -> None:
#         self._stopped.set()
#         self._runtime._message_queue.shutdown(immediate=True)  # type: ignore
#         await self._run_task

#     async def stop_when_idle(self) -> None:
#         await self._runtime._message_queue.join()  # type: ignore
#         self._stopped.set()
#         self._runtime._message_queue.shutdown(immediate=True)  # type: ignore
#         await self._run_task

#     async def stop_when(self, condition: Callable[[], bool], check_period: float = 1.0) -> None:
#         async def check_condition() -> None:
#             while not condition():
#                 await asyncio.sleep(check_period)
#             await self.stop()

#         await asyncio.create_task(check_condition())


from typing import Any, List, Mapping
import json
from autogen_core import AgentId,DefaultTopicId, MessageContext, event, rpc

from autogen_agentchat.messages import BaseAgentEvent, BaseChatMessage, MessageFactory

from autogen_agentchat.base import Response
from autogen_agentchat.state import ChatAgentContainerState
from autogen_agentchat.teams._group_chat._events import (
    GroupChatAgentResponse,
    GroupChatError,
    GroupChatMessage,
    GroupChatPause,
    GroupChatRequestPublish,
    GroupChatReset,
    GroupChatResume,
    GroupChatStart,
    SerializableException,
)
from drsai.modules.managers.messages.groupchat_messages import (
    GroupChatAgentLongTask,
    GroupChatLazyInit,
    GroupChatClose
    )
from drsai.modules.managers.messages.agent_messages import (
    AgentLongTaskMessage,
    LongTaskQueryMessage,
    )
from drsai.modules.baseagent.drsaiagent import DrSaiAgent
from drsai.utils.utils import construct_task
from autogen_agentchat.teams._group_chat._sequential_routed_agent import SequentialRoutedAgent
from loguru import logger

class DrSaiChatAgentContainer(SequentialRoutedAgent):
    """A core agent class that delegates message handling to an
    :class:`autogen_agentchat.base.ChatAgent` so that it can be used in a
    group chat team.

    Args:
        parent_topic_type (str): The topic type of the parent orchestrator.
        output_topic_type (str): The topic type for the output.
        agent (ChatAgent): The agent to delegate message handling to.
        message_factory (MessageFactory): The message factory to use for
            creating messages from JSON data.
    """

    def __init__(
        self,
        parent_topic_type: str,
        output_topic_type: str,
        agent: DrSaiAgent,
        message_factory: MessageFactory,
        long_task_topic_type: str | None = None,
    ) -> None:
        super().__init__(
            description=agent.description,
            sequential_message_types=[
                GroupChatStart,
                GroupChatRequestPublish,
                GroupChatReset,
                GroupChatAgentResponse,
                GroupChatAgentLongTask,
            ],
        )
        self._parent_topic_type = parent_topic_type
        self._output_topic_type = output_topic_type
        self._agent = agent
        self._message_buffer: List[BaseChatMessage] = []
        self._message_factory = message_factory
        # 专门用于长任务通信的 topic
        self._long_task_topic_type = long_task_topic_type or f"{parent_topic_type}_long_task"

    @rpc
    async def handle_lazy_init(self, message: GroupChatLazyInit, ctx: MessageContext) -> None:
        """Handle a lazy_init event by resetting the agent."""
        await self._agent.lazy_init(ctx.cancellation_token)

    @event
    async def handle_start(self, message: GroupChatStart, ctx: MessageContext) -> None:
        """Handle a start event by appending the content to the buffer."""
        if message.messages is not None:
            for msg in message.messages:
                self._buffer_message(msg)

    @event
    async def handle_agent_response(self, message: GroupChatAgentResponse, ctx: MessageContext) -> None:
        """Handle an agent response event by appending the content to the buffer."""
        chat_message = message.agent_response.chat_message
        if chat_message.source == "user_proxy" and \
            chat_message.metadata.get("attached_files"):
                query=chat_message.metadata.pop("content")
                files=json.loads(chat_message.metadata.pop("attached_files"))
                messages_return = construct_task(
                        query=query, 
                        files=files,
                        metadata=chat_message.metadata
                    )
                for msg in messages_return:
                    self._buffer_message(msg)
        else:
            self._buffer_message(message.agent_response.chat_message)
    
    @event
    async def handle_long_task(self, message: GroupChatAgentLongTask, ctx: MessageContext) -> None:
        """Handle Agent's long task."""
        cancellation_token = ctx.cancellation_token

        # Only support one long task query here
        try:
            request_message: LongTaskQueryMessage = message.message
            if isinstance(request_message, LongTaskQueryMessage):
                await self._log_message(request_message)
            else:
                raise ValueError("Invalid message type.")

            if request_message is None:
                raise ValueError("No long task query message found.")

            response: Response | None = None
            async for msg in self._agent._process_long_task_query(
                task=request_message,
                cancellation_token=cancellation_token,
            ):
                if isinstance(msg, Response):
                    response = msg
                    await self._log_message(msg.chat_message)
                    # if AgentLongTaskMessage, publish the message with long task to the long_task_topic
                    if isinstance(msg.chat_message, AgentLongTaskMessage):
                        await self.publish_message(
                            GroupChatAgentLongTask(message=msg.chat_message),
                            topic_id=DefaultTopicId(type=self._long_task_topic_type),
                            cancellation_token=cancellation_token,
                        )
                    else:
                        # if other general textmessage, publish the message to the group chat manager for next round
                        await self.publish_message(
                            GroupChatAgentResponse(agent_response=response),
                            topic_id=DefaultTopicId(type=self._parent_topic_type),
                            cancellation_token=ctx.cancellation_token,
                        )
                else:
                    await self._log_message(msg)

        except Exception as e:
            # Publish the error to the group chat.
            error_message = SerializableException.from_exception(e)
            await self.publish_message(
                GroupChatError(error=error_message),
                topic_id=DefaultTopicId(type=self._parent_topic_type),
                cancellation_token=ctx.cancellation_token,
            )
            # Raise the error to the runtime.
            raise

    @rpc
    async def handle_reset(self, message: GroupChatReset, ctx: MessageContext) -> None:
        """Handle a reset event by resetting the agent."""
        self._message_buffer.clear()
        await self._agent.on_reset(ctx.cancellation_token)

    @event
    async def handle_request(self, message: GroupChatRequestPublish, ctx: MessageContext) -> None:
        """Handle a content request event by passing the messages in the buffer
        to the delegate agent and publish the response."""
        try:
            # Pass the messages in the buffer to the delegate agent.
            response: Response | None = None
            async for msg in self._agent.on_messages_stream(self._message_buffer, ctx.cancellation_token):
                if isinstance(msg, Response):
                    await self._log_message(msg.chat_message)
                    response = msg
                else:
                    await self._log_message(msg)
            if response is None:
                raise ValueError(
                    "The agent did not produce a final response. Check the agent's on_messages_stream method."
                )
            # Publish the response to the group chat.
            self._message_buffer.clear()

            if msg.chat_message.source=="user_proxy":
                if msg.chat_message.metadata.get("user_request"):
                    msg.chat_message.content = msg.chat_message.metadata.get("content")

            if isinstance(response.chat_message, AgentLongTaskMessage):
                # 发布到专门的长任务 topic
                await self.publish_message(
                    GroupChatAgentLongTask(message=response.chat_message),
                    topic_id=DefaultTopicId(type=self._long_task_topic_type),
                    cancellation_token=ctx.cancellation_token,
                )
            else:
                await self.publish_message(
                    GroupChatAgentResponse(agent_response=response),
                    topic_id=DefaultTopicId(type=self._parent_topic_type),
                    cancellation_token=ctx.cancellation_token,
                )
        except Exception as e:
            # Publish the error to the group chat.
            error_message = SerializableException.from_exception(e)
            await self.publish_message(
                GroupChatError(error=error_message),
                topic_id=DefaultTopicId(type=self._parent_topic_type),
                cancellation_token=ctx.cancellation_token,
            )
            # Raise the error to the runtime.
            raise

    def _buffer_message(self, message: BaseChatMessage) -> None:
        if not self._message_factory.is_registered(message.__class__):
            raise ValueError(f"Message type {message.__class__} is not registered.")
        # Buffer the message.
        self._message_buffer.append(message)

    async def _log_message(self, message: BaseAgentEvent | BaseChatMessage) -> None:
        if not self._message_factory.is_registered(message.__class__):
            raise ValueError(f"Message type {message.__class__} is not registered.")
        # Log the message.
        await self.publish_message(
            GroupChatMessage(message=message),
            topic_id=DefaultTopicId(type=self._output_topic_type),
        )

    @rpc
    async def handle_pause(self, message: GroupChatPause, ctx: MessageContext) -> None:
        """Handle a pause event by pausing the agent."""
        if hasattr(self._agent, "pause"):
            await self._agent.pause(ctx.cancellation_token)
        else:
            await self._agent.on_pause(ctx.cancellation_token)

    @rpc
    async def handle_resume(self, message: GroupChatResume, ctx: MessageContext) -> None:
        """Handle a resume event by resuming the agent."""
        if hasattr(self._agent, "resume"):
            await self._agent.resume(ctx.cancellation_token)
        else:
            await self._agent.on_resume(ctx.cancellation_token)
    
    @rpc
    async def handle_close(self, message: GroupChatClose, ctx: MessageContext) -> None:
        """Handle a lazy_init event by resetting the agent."""
        await self._agent.close(ctx.cancellation_token)

    async def on_unhandled_message(self, message: Any, ctx: MessageContext) -> None:
        # raise ValueError(f"Unhandled message in agent container: {type(message)}")
        logger.warning(f"Unhandled message in agent container: {type(message)}")

    async def save_state(self) -> Mapping[str, Any]:
        agent_state = await self._agent.save_state()
        state = ChatAgentContainerState(
            agent_state=agent_state, message_buffer=[message.dump() for message in self._message_buffer]
        )
        return state.model_dump()

    async def load_state(self, state: Mapping[str, Any]) -> None:
        container_state = ChatAgentContainerState.model_validate(state)
        self._message_buffer = []
        for message_data in container_state.message_buffer:
            message = self._message_factory.create(message_data)
            if isinstance(message, BaseChatMessage):
                self._message_buffer.append(message)
            else:
                raise ValueError(f"Invalid message type in message buffer: {type(message)}")
        await self._agent.load_state(container_state.agent_state)



from typing import (
    AsyncGenerator, 
    List, 
    Sequence, 
    Dict, 
    Any, 
    Callable, 
    Awaitable, 
    Union, 
    Optional, 
    Tuple,
    Self,
    )

import asyncio
from loguru import logger
import inspect
import json
import os

from pydantic import BaseModel

from autogen_core import CancellationToken, FunctionCall
from autogen_core.tools import (
    BaseTool, 
    Workbench, 
    ToolSchema)
from autogen_core.memory import Memory
from autogen_core.model_context import ChatCompletionContext
from autogen_core.models import (
    ChatCompletionClient,
    CreateResult,
    FunctionExecutionResultMessage,
    FunctionExecutionResult,
    LLMMessage,
    UserMessage,
    AssistantMessage,
    SystemMessage,
    RequestUsage,
)

from autogen_agentchat.agents._assistant_agent import AssistantAgentConfig
from autogen_agentchat.base import Handoff as HandoffBase
from autogen_agentchat.base import Response
from autogen_agentchat.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    AgentEvent,
    ChatMessage,
    HandoffMessage,
    MemoryQueryEvent,
    ModelClientStreamingChunkEvent,
    TextMessage,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
    ToolCallSummaryMessage,
    UserInputRequestedEvent,
    ThoughtEvent,
    StructuredMessageFactory,
    # MultiModalMessage,
    Image,
)

from .drsaiagent import DrSaiAgent, DrSaiAgentConfig
from drsai.modules.managers.messages.agent_messages import(
    AgentLongTaskMessage,
    LongTaskQueryMessage,
    AgentLogEvent,
    ToolLongTaskEvent,
)
from drsai.modules.components.task_manager.base_task_system import TaskStatus
from drsai.modules.managers.database import DatabaseManager

class LongTaskAgentConfig(DrSaiAgentConfig):
    """
    A configuration class for LongTaskAgent.
    """
    tool_name: str|None = None
    tool_status: str|None = None


class LongTaskAgent(DrSaiAgent):
    """
    # TODO: 一个实现可以执行长时间任务执行的和查询的智能体
    
    """
    def __init__(
        self,
        name: str,
        *,
        model_client: ChatCompletionClient = None,
        tools: List[BaseTool[Any, Any] | Callable[..., Any] | Callable[..., Awaitable[Any]]] | None = None,
        workbench: Workbench | None = None,
        handoffs: List[HandoffBase | str] | None = None,
        model_context: ChatCompletionContext | None = None,
        description: str = "An agent that provides assistance with ability to use tools.",
        system_message: (
            str | None
        ) = "You are a helpful AI assistant. Solve tasks using your tools. Reply with TERMINATE when the task has been completed.",
        model_client_stream: bool = True,
        reflect_on_tool_use: bool | None = None,
        tool_call_summary_format: str = "{result}",
        output_content_type: type[BaseModel] | None = None,
        output_content_type_format: str | None = None,
        memory: Sequence[Memory] | None = None,
        metadata: Dict[str, str] | None = None,

        # drsaiAgent specific
        memory_function: Callable = None,
        # allow_reply_function: bool = False,
        reply_function: Callable = None,
        db_manager: DatabaseManager = None,
        thread_id: str = None,
        user_id: str = None,
        **kwargs,
            ):
        
        super().__init__(
            name, 
            model_client,
            tools=tools,
            workbench=workbench,
            handoffs=handoffs,
            model_context=model_context,
            description=description,
            system_message=system_message,
            model_client_stream=model_client_stream,
            reflect_on_tool_use=reflect_on_tool_use,
            tool_call_summary_format=tool_call_summary_format,
            output_content_type=output_content_type,
            output_content_type_format=output_content_type_format,
            memory=memory,
            metadata=metadata,
            memory_function=memory_function,
            reply_function=reply_function,
            db_manager=db_manager,
            thread_id=thread_id,
            user_id=user_id,
            **kwargs,
            )
    
        self._tool_name: Optional[str] = None
        self._tool_arguments: Optional[Dict[str, Any]] = None
    
    async def on_messages_stream(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """
        Process the incoming messages with the assistant agent and yield events/responses as they happen.
        """

        # monitor the pause event
        if self.is_paused:
            yield Response(
                chat_message=TextMessage(
                    content=f"The {self.name} is paused.",
                    source=self.name,
                    metadata={"internal": "yes"},
                )
            )
            return

        # Set up background task to monitor the pause event and cancel the task if paused.
        async def monitor_pause() -> None:
            await self._paused.wait()
            self.is_paused = True

        monitor_pause_task = asyncio.create_task(monitor_pause())
        inner_messages: List[BaseAgentEvent | BaseChatMessage] = []
        try:
            # Gather all relevant state here
            agent_name = self.name
            model_context = self._model_context
            memory = self._memory
            system_messages = self._system_messages
            workbench = self._workbench
            handoff_tools = self._handoff_tools
            handoffs = self._handoffs
            model_client = self._model_client
            model_client_stream = self._model_client_stream
            reflect_on_tool_use = self._reflect_on_tool_use
            tool_call_summary_format = self._tool_call_summary_format
            output_content_type = self._output_content_type
            format_string = self._output_content_type_format

            # STEP 1: Add new user/handoff messages to the model context
            await self._add_messages_to_context(
                model_context=model_context,
                messages=messages,
            )
            
            # STEP 2: Update model context with any relevant memory
            for event_msg in await self._update_model_context_with_memory(
                memory=memory,
                model_context=model_context,
                agent_name=agent_name,
            ):
                inner_messages.append(event_msg)
                yield event_msg

            # STEP 3: Run the first inference
            model_result = None
            async for inference_output in self._call_llm(
                model_client=model_client,
                model_client_stream=model_client_stream,
                system_messages=system_messages,
                model_context=model_context,
                workbench=workbench,
                handoff_tools=handoff_tools,
                agent_name=agent_name,
                cancellation_token=cancellation_token,
                output_content_type=output_content_type,
            ):
                if self.is_paused:
                    raise asyncio.CancelledError()
                
                if isinstance(inference_output, CreateResult):
                    model_result = inference_output
                else:
                    # Streaming chunk event
                    yield inference_output

            assert model_result is not None, "No model result was produced."

            # --- NEW: If the model produced a hidden "thought," yield it as an event ---
            if model_result.thought:
                thought_event = ThoughtEvent(content=model_result.thought, source=agent_name)
                yield thought_event
                inner_messages.append(thought_event)

            # Add the assistant message to the model context (including thought if present)
            await model_context.add_message(
                AssistantMessage(
                    content=model_result.content,
                    source=agent_name,
                    thought=getattr(model_result, "thought", None),
                )
            )

            # For long task 
            if not isinstance(model_result.content, str):
                self._tool_name = model_result.content[0].name
                self._tool_arguments = json.loads(model_result.content[0].arguments)
                
            # STEP 4: Process the model output
            async for output_event in self._process_model_result(
                model_result=model_result,
                inner_messages=inner_messages,
                cancellation_token=cancellation_token,
                agent_name=agent_name,
                system_messages=system_messages,
                model_context=model_context,
                workbench=workbench,
                handoff_tools=handoff_tools,
                handoffs=handoffs,
                model_client=model_client,
                model_client_stream=model_client_stream,
                reflect_on_tool_use=reflect_on_tool_use,
                tool_call_summary_format=tool_call_summary_format,
                output_content_type=output_content_type,
                format_string=format_string,
            ): 
                if isinstance(output_event, Response):
                    if isinstance(output_event.chat_message, ToolCallSummaryMessage):
                        mcp_output = json.loads(output_event.chat_message.content)
                        if mcp_output["status"] == "IN_PROGRESS":
                            self._tool_arguments["task_id"] = mcp_output["id"]
                            output_event.chat_message =  AgentLongTaskMessage(
                                source=self.name,
                                content=f"{self._tool_name} is running. Please wait for the result.",
                                task_status=TaskStatus.in_progress.value,
                                query_arguments=self._tool_arguments,
                                tool_name=self._tool_name
                            )
                        yield output_event
                else:
                    yield output_event

        except asyncio.CancelledError:
            # If the task is cancelled, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The task was cancelled by the user.",
                    source=self.name,
                    metadata={"internal": "yes"},
                ),
                inner_messages=inner_messages,
            )
        except Exception as e:
            logger.error(f"Error in {self.name}: {e}")
            # add to chat history
            await model_context.add_message(
                AssistantMessage(
                    content=f"An error occurred while executing the task: {e}",
                    source=self.name
                )
            )
            yield Response(
                chat_message=TextMessage(
                    content=f"An error occurred while executing the task: {e}",
                    source=self.name,
                    metadata={"internal": "no"},
                ),
                inner_messages=inner_messages,
            )
        finally:
            # Cancel the monitor task.
            try:
                monitor_pause_task.cancel()
                await monitor_pause_task
            except asyncio.CancelledError:
                pass

    async def _process_long_task_query(
            self,
            task: Dict|LongTaskQueryMessage|Sequence[BaseChatMessage] | None = None,
            cancellation_token: CancellationToken | None = None,
        )-> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        
        if not isinstance(task, LongTaskQueryMessage):
            raise ValueError("tasks must be a LongTaskQueryMessage")
        
        query_arguments: Dict = task.query_arguments
        query_tool_name = task.tool_name
        if query_tool_name is None or query_arguments is None:
            raise ValueError("query_tool_name or query_arguments cannot be None")
        
        result = await self._workbench.call_tool(
                name = query_tool_name,
                arguments=query_arguments,
                cancellation_token=cancellation_token
                )
        result_json = json.loads(result.result[0].content)

        if result_json["status"] == "IN_PROGRESS":
            yield Response(
                chat_message=AgentLongTaskMessage(
                    source=self.name,
                    content=result_json["result"],
                    task_status=TaskStatus.in_progress.value,
                    query_arguments=query_arguments,
                    tool_name=query_tool_name
                ))
        else:
            yield Response(
                chat_message=TextMessage(
                        source=self.name,
                        content=result_json["result"],
                    ))


'''
A Model_Context component that keep agent's memory using RAGFlow and LLM
'''
from pydantic import BaseModel, Field, field_serializer
from typing import (
    List, 
    Dict, 
    Any,
    Self,
    Literal,
    Optional
)

from autogen_core import(
    Component, 
    ComponentModel
)
from autogen_core.tools import ( 
    ToolSchema,
    ParametersSchema,
)
from autogen_core import CancellationToken
from autogen_core.models import (
    ChatCompletionClient,
    FunctionExecutionResultMessage,
    LLMMessage,
    UserMessage,
    AssistantMessage,
    FunctionExecutionResult,
    SystemMessage,
)
from autogen_core.model_context import ChatCompletionContext

from drsai.modules.components.memory.ragflow_memory import RAGFlowMemoryManager

from loguru import logger
from datetime import datetime
import json
import os
from datetime import datetime
from dataclasses import asdict

class LocalMesssage(BaseModel):
    """
    A message in a local memory.
    """
    role: Literal["user", "assistant", "system"]
    content: str
    source: str | None = None
    is_tool_call: bool = False # json.dumps(tool_call)->content
    create_time: datetime = Field(default_factory=datetime.now)

    @field_serializer('create_time')
    def serialize_dt(self, dt: datetime) -> str:
        return dt.strftime('%Y-%m-%d %H:%M:%S')

COMPRESSION_PROMPT_ZN = """
你是一个负责压缩长对话记忆的助手。现在给你一段包含用户、智能助手{name}以及其他助手多轮对话的记录。你的任务是从中提取长期有价值的信息，并输出高度压缩、结构清晰的摘要。

请从以下方面进行总结：

1. **用户的初始任务、目标、背景信息、用户或者其他智能助手查询的记忆等信息**（只保留长期有用内容）。
2. **智能助手{name}在对话中做出的关键回复、思考过程摘要和重要决策**：
   - 包含每个关键动作使用的工具名称、调用目的、输入要点、输出或执行结果。
   - 若工具失败或返回无效结果，也需简要记录原因。
3. **当前轮对话结束时，用户或其他助手对{name}提出的最新需求或待办事项**。

同时请遵守以下要求：
- 忽略闲聊、重复确认、无用解释等短期内容。
- 若对话是中英文或混合语言，保留相应的语言。
- 尽可能压缩 token，同时保持可读性与完整的因果链。
- 输出必须客观、无臆测，仅基于对话内容进行总结。

请最终按照每个部分进行输出。

"""

COMPRESSION_PROMPT_EN = """
You are an assistant responsible for compressing long multi-agent conversations into concise, long-term memory summaries. You will receive a conversation containing the user, assistant {name}, and possibly other agents. Your task is to extract only the high-value, long-term information and produce a highly compressed and structured summary.

Please summarize the conversation from the following perspectives:

1. **The user’s initial goal, task description, and any relevant background information** (keep only information that remains useful later).
2. **Key actions, reasoning steps, and major decisions made by assistant {name}:**
   - Include the tools used, the purpose of each tool call, the main inputs, and the outputs or results.
   - If a tool call failed or returned an invalid result, briefly record the reason.
3. **The current outstanding request or the latest requirement from the user or other agents toward assistant {name}.**

Additional instructions:
- Ignore small talk, repeated confirmations, and temporary or low-value content.
- The input may contain Chinese, English, or mixed languages; keep the origin languages**, clearly structured.
- Compress aggressively to minimize token usage while keeping essential causal chains.
- Do not infer or invent any information that is not explicitly stated.

Output the final result as a structured bullet-point summary.
"""

def get_judgment_tool(strict: bool = False,) -> ToolSchema:
    parameters = ParametersSchema(
        type="object",
        properties={
            "be_documented": {
                "type": "boolean",
                "description":  "The conversation should be documented or not."
            }
        },
        required=["be_documented"],
        additionalProperties=False,
    )
    tool_schema = ToolSchema(
        name="judgment_documented",
        description=f"Determine whether the aforementioned conversation should be recorded",
        parameters=parameters,
        strict=strict,
    )
    return tool_schema

SUMMARY_PROMPT_EN = """You need to analyze the following dialogue and summarize them in bullet points according to the requirements below:

- What is the user's task?
- How does the intelligent assistant: {name} respond to user questions? What tools/skills are used, and what steps are taken?
- What errors occurred in the middle, were they corrected through feedback, how were they corrected, and what were the results of the correction?
- What kind of reply did the intelligent assistant: {name} finally give?

If the user's question is merely a greeting or an inquiry about identity, call the `judgment_documented` tool to respond with false, indicating that no record is needed.

**NOTE:**

- If documenting is need, do not call the `judgment_documented` tool.

"""

class DrSaiChatCompletionContextConfig(BaseModel):
    agent_name: str
    model_client: ComponentModel
    user_id: str
    thread_id: str
    work_dir: str
    token_limit: int | None = None
    compression_prompt: str|None = None
    summary_task_prompt: str|None = None
    rag_flow_url: str | None = None
    rag_flow_token: str | None = None
    dataset_id: str|None = None,
    document_id: str|None = None,
    tool_schema: List[ToolSchema] | None = None
    learning_dataset_id: str|None = None
    learning_document_id: str|None = None
    initial_messages: List[LLMMessage] | None = None
    history_messages: List[LLMMessage] | None = None

class DrSaiChatCompletionContext(ChatCompletionContext, Component[DrSaiChatCompletionContextConfig]):
    """
    A context that limits the number of tokens using LLM and store memory using RAGFlow.
    Note:
        - This ChatCompletionContext keeps the first two SystemMessage and the last two UserMessage.
    """

    component_config_schema = DrSaiChatCompletionContextConfig
    component_provider_override = "drsai.DrSaiChatCompletionContext"
    component_description = "A context that limits the number of tokens used by the model."

    def __init__(
        self,
        agent_name: str,
        model_client: ChatCompletionClient,
        *,
        user_id: str | None = None,
        thread_id: str | None = None,
        work_dir: str | None = None, 
        token_limit: int | None = None,
        compression_prompt: str|None = None,
        summary_task_prompt: str|None = None,
        rag_flow_url: str | None = None,
        rag_flow_token: str | None = None,
        dataset_id: str|None = None,
        document_id: str|None = None,
        learning_dataset_id: str|None = None,
        learning_document_id: str|None = None,
        # upload_memory: bool = False,
        tool_schema: List[ToolSchema] = [],
        initial_messages: List[LLMMessage] | None = None,
    ) -> None:
        """
        agent_name: The name of the agent.
        model_client: The model client to use.
        token_limit: The maximum number of tokens to use.
        compression_prompt: The prompt to use for compressing the conversation.
        rag_flow_url: The url of ragflow.
        rag_flow_token: The token of ragflow.
        // upload_memory: update memory to ragflow when add_message is called.
        dataset_id: The id of dataset for memory storage.
        document_id: The id of document for memory storage. 
        tool_schema: The schema of tools.
        initial_messages: The initial messages to use.
        """

        super().__init__(initial_messages)

        self._agent_name = agent_name
        if token_limit is not None and token_limit <= 0:
            raise ValueError("token_limit must be greater than 0.")
        self._token_limit = token_limit
        self._model_client = model_client

        self._thread_id = thread_id
        self._user_id = user_id
        self._work_dir = work_dir

        self._compression_prompt = compression_prompt or COMPRESSION_PROMPT_EN.format(name=agent_name)
        self._summary_task_prompt = summary_task_prompt or SUMMARY_PROMPT_EN.format(name=agent_name)

        self._dataset_id = dataset_id
        self._document_id = document_id
        self._learning_dataset_id = learning_dataset_id
        self._learning_document_id = learning_document_id

        self._rag_flow_manager = None
        if rag_flow_url is not None and rag_flow_token is not None:
            self._rag_flow_manager = RAGFlowMemoryManager(rag_flow_url, rag_flow_token)
            if dataset_id is None:
                raise ValueError("dataset_id must be provided when rag_flow_url and rag_flow_token are provided.")
        
        # self._is_upload_memory = upload_memory 

        self._tool_schema = tool_schema

        self._history_messages: list[dict] = []

        self._current_messages: list[LLMMessage] = []

    async def add_message(
            self, 
            message: LLMMessage, 
            # important_keywords: List[str] = None,
            # questions: List[str] = None,
            # document_id: str = None
            ) -> None:
        """
        Add a message to the context and store the content to ragflow memory manager.
        important_keywords: The key terms or phrases to tag with the chunk.
        questions: If there is a given question, the embedded chunks will be based on them.
        """

        # if self._rag_flow_manager is not None and self._is_upload_memory:
        #     await self._rag_flow_manager.add_chunks_to_dataset(
        #         dataset_id = self._dataset_id,
        #         document_id = document_id or self._document_id,
        #         content = message.content,
        #         important_keywords = important_keywords,
        #         questions = questions)
        self._messages.append(message)
        self._current_messages.append(message)

        if isinstance(message, SystemMessage):
            local_messages = LocalMesssage(
                role="system",
                content=message.content
            )
        elif isinstance(message, UserMessage):
            local_messages = LocalMesssage(
                role="user",
                content=message.content,
                name=message.source
            )
        elif isinstance(message, AssistantMessage):
            content = message.content
            if isinstance(message.content, list):
                # content = json.dumps([asdict(tool_call) for tool_call in message.content])
                content = str(message.content)
            local_messages = LocalMesssage(
                role="assistant",
                content=content,
                name=message.source
            )
        elif isinstance(message, FunctionExecutionResultMessage): 
            tool_call_list = [tool_call.model_dump() for tool_call in message.content]
            local_messages = LocalMesssage(
                role="assistant",
                content=json.dumps(tool_call_list),
                is_tool_call=True
            )
        self._history_messages.append(local_messages.model_dump())

    def format_messages_str(self, messages: List[LLMMessage]) -> str:
        """Summarize current messages."""
        content_str = "The conversations:\n\n"
        for message in messages:
            if isinstance(message, SystemMessage):
                content_str += f"System: {message.content}\n"
            elif isinstance(message, UserMessage):
                content_str += f"{message.source}: {message.content}\n"
            elif isinstance(message, AssistantMessage):
                content_str += f"{message.source}: {message.content}\n"
            elif isinstance(message, FunctionExecutionResultMessage):
                content_str += f"Tool Calling: {str(message.content)}\n"
        return content_str
    
    async def get_messages(
            self, 
            cancellation_token: CancellationToken = None) -> List[LLMMessage]:
        """Get at most `token_limit` tokens in recent messages. If the token limit is not
        provided, then return as many messages as the remaining token allowed by the model client."""
        messages = list(self._messages)
       
        try:
            if self._token_limit is not None:
                token_count = self._model_client.count_tokens(messages, tools=self._tool_schema)
                if token_count > self._token_limit:
                    messages_str = self.format_messages_str(messages)
                    compressed_response = await self._model_client.create(
                        messages=[
                            UserMessage(source="user", content=self._compression_prompt+messages_str)
                        ],
                        cancellation_token=cancellation_token,
                    )
                    compressed_content = compressed_response.content
                    
                    # keep some key messages
                    remaining_messages = []
                    
                    # Keep first two SystemMessage
                    system_count = 0
                    for message in messages:
                        if isinstance(message, SystemMessage) and system_count < 2:
                            remaining_messages.append(message)
                            system_count += 1
                    
                    # Add compressed content as UserMessage in the middle
                    remaining_messages.append(UserMessage(content=compressed_content, source="system"))
                    
                    # Keep last UserMessage
                    user_messages = [msg for msg in messages if isinstance(msg, UserMessage)]
                    for user_message in reversed(user_messages):
                        if user_message.source == "user":
                            remaining_messages.append(user_message)
                            break
                    
                    # Update messages
                    self._messages = remaining_messages
                    return remaining_messages
                else:
                    return messages
            else:
                raise ValueError("token_limit must be provided.")

        except Exception as e:
            logger.error(f"There is an error when compressing the memory using LLM: {e}. We have to truncate the memory.")
            if self._token_limit is None:
                remaining_tokens = self._model_client.remaining_tokens(messages, tools=self._tool_schema)
                while remaining_tokens < 0 and len(messages) > 0:
                    middle_index = len(messages) // 2
                    messages.pop(middle_index)
                    remaining_tokens = self._model_client.remaining_tokens(messages, tools=self._tool_schema)
            else:
                token_count = self._model_client.count_tokens(messages, tools=self._tool_schema)
                while token_count > self._token_limit and len(messages) > 0:
                    middle_index = len(messages) // 2
                    messages.pop(middle_index)
                    token_count = self._model_client.count_tokens(messages, tools=self._tool_schema)
            if messages and isinstance(messages[0], FunctionExecutionResultMessage):
                # Handle the first message is a function call result message.
                # Remove the first message from the list.
                messages = messages[1:]
            # Update messages
            self._messages = messages
            return messages

    async def create_new_session_document(
        self,
        user_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        thread_id: Optional[str] = None,
        work_dir: Optional[str] = None,
        create_type: str = "session" # or learning_memory
        ) -> str:
        """
        Create a temporary file named {user_id}_{thread_id}.txt in work_dir,
        upload it to RAGFlow, then attach user_id and thread_id as meta_fields.

        Returns:
            The document_id of the newly created document.
        """
        if self._rag_flow_manager is None:
            raise ValueError("RAGFlow manager is not configured.")
        
        user_id = user_id or self._user_id
        dataset_id = dataset_id or self._dataset_id
        thread_id = thread_id or self._thread_id
        work_dir = work_dir or self._work_dir

        # 1. cerate session document
        if create_type == "learning_memory":
            file_name = f"{user_id}_learning_memory.txt"
            meta_fields = {"user_id": user_id, "learning_dataset_id": dataset_id}
        else:
            file_name = f"{user_id}_{thread_id}.txt"
            meta_fields={"user_id": user_id, "thread_id": thread_id},
        file_path = os.path.join(work_dir, file_name)
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(f"user_id: {user_id}\nthread_id: {thread_id}\n")

        try:
            # 2. upload to RAGFlow and get document_id
            document_ids = await self._rag_flow_manager.add_files_to_dataset_and_parse(
                dataset_id=dataset_id,
                files_path=file_path,
            )
            document_id = document_ids[0]

            # 3. write user_id / thread_id to document meta
            await self._rag_flow_manager.update_document(
                dataset_id=dataset_id,
                document_id=document_id,
                meta_fields=meta_fields,
            )
        finally:
            # 4. delete the temporary file
            if os.path.exists(file_path):
                os.remove(file_path)

        if create_type == "learning_memory":
            self._learning_document_id = document_id
        else:
            self._document_id = document_id

        logger.info(f"Session document: {document_id} created for user_id: {user_id}, thread_id: {thread_id}")

        return document_id

    async def upload_conversation_to_ragflow(
            self, 
            current_messages: Optional[List[LLMMessage]] = None,
            document_id: Optional[str] = None,
            important_keywords: Optional[List[str]] = None,
            questions: Optional[List[str]] = None
            ) -> str:
        """Upload the conversation to RAGFlow."""

        now_str = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        important_keywords = important_keywords or [f"thread_id:{self._thread_id}", f"timestamp:{now_str}"]
        
        current_messages = current_messages or self._current_messages
        if self._rag_flow_manager:
            for message in current_messages:
                if isinstance(message, UserMessage):
                    questions = [message.content]
                    continue
                await self._rag_flow_manager.add_chunks_to_dataset(
                    dataset_id = self._dataset_id,
                    document_id = document_id or self._document_id,
                    content = str(message.content),
                    important_keywords = important_keywords,
                    questions = questions)

    async def summry_conversation_to_ragflow(
            self, 
            current_messages: Optional[List[LLMMessage]] = None,
            summary_task_prompt: Optional[str] = None,
            dataset_id: Optional[str] = None,
            document_id: Optional[str] = None,
            thread_id: Optional[str] = None,
            cancellation_token: Optional[CancellationToken] = None,
            ) -> str:
        """Summarize the conversation."""
        
        if self._rag_flow_manager is None:
            raise ValueError("RAGFlow manager is not configured.")
        current_messages = current_messages or self._current_messages
        questions = [message.content for message in current_messages if message.source == "user" or message.source == "user_proxy"]
        summry_prompt = summary_task_prompt or self._summary_task_prompt
        messages_str = self.format_messages_str(current_messages)
        summry_response = await self._model_client.create(
            messages=[
                UserMessage(source="user", content=summry_prompt+messages_str)
            ],
            cancellation_token=cancellation_token,
        )
        summry_content = summry_response.content
        # judgment_tools = [get_judgment_tool()]
        # messages_str = self.format_messages_str(current_messages)
        # summry_response = await self._model_client.create(
        #     messages=[
        #         UserMessage(source="user", content=summry_prompt+messages_str)
        #     ],
        #     tools=judgment_tools,
        #     cancellation_token=cancellation_token,
        # )
        # summry_content = summry_response.content

        # thread_id = thread_id or self._thread_id

        # if isinstance(summry_content, list):
        #     arguments = json.loads(summry_content[0].arguments)
        #     if arguments.get("be_documented", False):
        #         summry_response = await self._model_client.create(
        #             messages=[
        #                 UserMessage(source="user", content=summry_prompt+messages_str)
        #             ],
        #             cancellation_token=cancellation_token,
        #         )
        #         summry_content = summry_response.content
        #     else:
        #         return "The conversation is not summarized."
        
        thread_id = thread_id or self._thread_id

        await self._rag_flow_manager.add_chunks_to_dataset(
            dataset_id = dataset_id,
            document_id = document_id,
            content = summry_content,
            questions = questions,
            important_keywords = [f"thread_id:{thread_id}"]
            )
        logger.info(f"Summarized conversation added to RAGFlow for thread_id: {thread_id}")
        return str(summry_content)
    
    async def summry_conversation_to_MEMORY_MD(
            self, 
            current_messages: Optional[List[LLMMessage]] = None,
            summary_task_prompt: Optional[str] = None,
            thread_id: Optional[str] = None,
            cancellation_token: Optional[CancellationToken] = None,
            ) -> str:
        """Summarize the conversation."""
        # TODO: 让AI读取MEMORY.md，进行近期记忆的刷新
        if self._rag_flow_manager is None:
            raise ValueError("RAGFlow manager is not configured.")
        current_messages = current_messages or self._current_messages
        summry_prompt = summary_task_prompt or self._summary_task_prompt
        messages_str = self.format_messages_str(current_messages)
        summry_response = await self._model_client.create(
            messages=[
                UserMessage(source="user", content=summry_prompt+messages_str)
            ],
            cancellation_token=cancellation_token,
        )
        summry_content = summry_response.content
       
        return str(summry_content)
    
    async def summry_conversation_to_memory(
            self,
            summary_task_prompt: Optional[str] = None
            ) -> None:
        """Summarize the conversation by user's prompt.
        
        Arguments:
            summary_task_prompt: The user's request for summarizing the conversation.
        
        Note:
            If the specific summarizing request is not provided , the default prompt will be used.
        """

        try:
            summry_content = await self.summry_conversation_to_ragflow(
                summary_task_prompt=summary_task_prompt,
                dataset_id=self._learning_dataset_id,
                document_id=self._learning_document_id
                )
            return f"Summarized conversation added to RAGFlow, The summry_content is: \n{summry_content}"
        except Exception as e:
            logger.error(f"Failed to summarize conversation: {e}")
            return "Failed to summarize conversation"

    def _extract_thread_id(self, chunk: Dict[str, Any]) -> Optional[str]:
        """ extract the type of 'thread_id:...' from the chunk"""
        for kw in chunk.get('important_keywords', []):
            if isinstance(kw, str) and kw.startswith('thread_id:'):
                return kw[10:]
        return "null"

    def _extract_timestamp(self, chunk: Dict[str, Any]) -> Optional[str]:
        """ extract the type of 'timestamp:...' from the chunk"""
        for kw in chunk.get('important_keywords', []):
            if isinstance(kw, str) and kw.startswith('timestamp:'):
                return kw[10:]
        return "null"

    async def retrieve_from_memory(
        self, 
        question: str,
        document_ids: List[str] = [],
        page_size: int = 10,
        similarity_threshold: float = 0.2,
        vector_similarity_weight: float = 0.3,
        top_k: int = 1024,
        rerank_id: str = "hepai/bge-reranker-v2-m3___OpenAI-API@OpenAI-API-Compatible",
        keyword: bool = True,
        cross_languages: list[str] = ["English", "Chinese"],
        metadata_condition: Optional[Dict[str, str]] = None,
        # user_id: Optional[str] = None,
        # thread_id: Optional[str] = None,
        ) -> str:
        """Retrieve relevant information from memory. No special requirements, just input the question.
        
        Arguments:
            question: The question to retrieve relevant information.
            document_ids: The document_ids to retrieve relevant information. Default is [].
            page_size: The number of chunks to retrieve. Default is 10.
            similarity_threshold: The similarity threshold. Default is 0.2.
            vector_similarity_weight: The vector similarity weight. Default is 0.3.
            top_k: The top_k. Default is 1024.
            rerank_id: The rerank_id. Default is "hepai/bge-reranker-v2-m3___OpenAI-API@OpenAI-API-Compatible".
            keyword: Whether to use keyword. Default is True.
            cross_languages: The cross_languages. Default is ["English", "Chinese"].
        
        Return:
                The string of relevant information.
        """
        kwargs: Dict[str, Any] = dict(
            question=question,
            dataset_ids=[self._dataset_id],
            similarity_threshold=similarity_threshold,
            vector_similarity_weight=vector_similarity_weight,
            page_size=page_size,
            top_k=top_k,
            rerank_id=rerank_id,
            keyword=keyword,
            cross_languages=cross_languages,
        )

        if document_ids:
            kwargs["document_ids"] = document_ids
            
        if metadata_condition:
            kwargs["metadata_condition"] = metadata_condition
        else:
            # user_id = user_id or self._user_id
            # thread_id = thread_id or self._thread_id , "thread_id": self._thread_id
            kwargs["metadata_condition"] = {"user_id": self._user_id}
        
        raw = await self._rag_flow_manager.retrieve_chunks_by_content(**kwargs)
        chunks = raw.get('chunks', []) if raw else []
        for chunk in chunks:
            chunk["thread_id"] = self._extract_thread_id(chunk)
            chunk["timestamp"] = self._extract_timestamp(chunk)

        if len(chunks) > 0:
            return "**Relevant information:** \n" + "\n".join(
                [f'Session_ID: {chunk["thread_id"]} | Time: {chunk["timestamp"]}\n{chunk["content"]}' for chunk in chunks]
            )
        else:
            return "No relevant information found."

    def _to_config(self) -> DrSaiChatCompletionContextConfig:
        return DrSaiChatCompletionContextConfig(
            model_client=self._model_client.dump_component(),
            token_limit=self._token_limit,
            compression_prompt=self._compression_prompt,
            rag_flow_url=self._rag_flow_manager.base_url if self._rag_flow_manager else None,
            rag_flow_token=self._rag_flow_manager.api_key if self._rag_flow_manager else None,
            dataset_id=self._dataset_id,
            document_id=self._document_id,
            tool_schema=self._tool_schema,
            initial_messages=self._initial_messages,
            history_messages=self._history_messages,
        )

    @classmethod
    def _from_config(cls, config: DrSaiChatCompletionContextConfig) -> Self:
        return cls(
            model_client=ChatCompletionClient.load_component(config.model_client),
            token_limit=config.token_limit,
            compression_prompt=config.compression_prompt,
            rag_flow_url=config.rag_flow_url,
            rag_flow_token=config.rag_flow_token,
            dataset_id=config.dataset_id,
            document_id=config.document_id,
            tool_schema=config.tool_schema,
            initial_messages=config.initial_messages,
        )





from typing import (
    AsyncGenerator, 
    List, 
    Sequence, 
    Dict, 
    Any, 
    Callable, 
    Awaitable, 
    Union, 
    Optional, 
    Tuple,
    Self,
    Mapping,
    )
import os, json, sys, uuid, shutil
import asyncio, traceback
from pydantic import BaseModel
from pathlib import Path
from loguru import logger

from drsai import CancellationToken, FunctionCall
from drsai.modules.baseagent import (
    DrSaiAgent, 
    HandoffBase, 
    Response, 
    TaskResult, 
    CreateResult,
    FunctionExecutionResultMessage,
    FunctionExecutionResult,
    AssistantMessage,
    UserMessage,
    SystemMessage,
    LLMMessage
    )
from drsai.modules.baseagent.drsaiagent import DrSaiAgentConfig
from drsai.modules.baseagent import CodeExecutorAgent, CodeExecutor
from ..drsai_worker_agent import HepAIWorkerAgent
from drsai.modules.components import (
    ComponentModel,
)
from drsai.modules.components.model_client import ChatCompletionClient, HepAIChatCompletionClient
from drsai.modules.components.model_context import (
    ChatCompletionContext,
    DrSaiChatCompletionContext,
)
from drsai.modules.components.memory import Memory
from drsai.modules.components.tool import (
    BaseTool, 
    FunctionTool, 
    Workbench,
    ToolSchema,
    ParametersSchema,
    StdioServerParams,
    SseServerParams,
    mcp_server_tools
    )
from drsai.modules.managers.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    AgentEvent,
    ChatMessage,
    HandoffMessage,
    MemoryQueryEvent,
    ModelClientStreamingChunkEvent,
    TextMessage,
    StopMessage,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
    ToolCallSummaryMessage,
    UserInputRequestedEvent,
    ThoughtEvent,
    AgentLogEvent,
    StructuredMessage,
    StructuredMessageFactory,
    MultiModalMessage,
    Image,
    
)
from drsai.modules.managers.database import DatabaseManager
from drsai.configs.constant import RUNS_DIR
from .managers import (
    UserProfileManager,
    TodoManager,
    get_operator_funcs,
)
from drsai.modules.components.skills import SkillLoader
from drsai.utils.utils import download_file_from_url_or_base64
from .managers.get_managers_tools import (
    get_agent_skills_tool,
    get_subagent_tools,
    get_todo_manager_tool,
    create_local_venv,
)

class DrSaiAssistantConfig(DrSaiAgentConfig):
    skills_dir: Optional[str | List[str]]
    work_dir: str | None
    only_in_workspace: bool
    extra_work_dirs: List[str]
    executor: ComponentModel
    sub_agent_config: Dict
    max_turn_count: int
    token_limit: int
    rag_flow_url: str
    rag_flow_token: str
    memory_dataset_id: str
    learning_dataset_id: str
    

class DrSaiAssistant(DrSaiAgent):
    """
    专业科学数据智能分析智能体

    TODO: 核心能力:
    1. 用户个人画像/提示词/其他配置的动态更新
    2. 任务规划与分解
    3. 多任务进度管理
    3. 智能工具/Skills/子智能体调用与动态更新
    4. 记忆注入机制与长期记忆
    5. 自我被动学习与skill转化
    6. 错误处理与用户交互
    """

    def __init__(
        self,
        name: str,
        *,
        model_client: ChatCompletionClient = None,
        tools: List[BaseTool[Any, Any] | Callable[..., Any] | Callable[..., Awaitable[Any]]] | None = None,
        workbench: Workbench | None = None,
        handoffs: List[HandoffBase | str] | None = None,
        model_context: ChatCompletionContext | None = None,
        description: str = "An agent that provides assistance with ability to use tools.",
        system_message: (
            str | None
        ) = "You are a helpful AI assistant. Solve tasks using your tools. Reply with TERMINATE when the task has been completed.",
        model_client_stream: bool = True,
        reflect_on_tool_use: bool | None = None,
        tool_call_summary_format: str = "{result}",
        tool_call_summary_prompt: str| None = None,
        output_content_type: type[BaseModel] | None = None,
        output_content_type_format: str | None = None,
        memory: Sequence[Memory] | None = None,
        metadata: Dict[str, str] | None = None,
        # drsaiAgent specific
        memory_function: Callable | None = None,
        reply_function: Callable | None = None,
        db_manager: DatabaseManager | None = None,
        thread_id: str | None = None,
        user_id: str | None = None,
        set_model_client: Callable | None = None,
        llm_mode_config: Dict = {},
        # skills and executor
        skills_dir: Optional[str | List[str]] = None,
        work_dir: str | None = None,
        only_in_workspace: bool = True,
        extra_work_dirs: List[str] | None = None,
        executor: CodeExecutor | None = None,
        sub_agent_config: Dict = {},
        max_turn_count: int = 20,
        token_limit: int = 50000,
        rag_flow_url: str | None = None,
        rag_flow_token: str | None = None,
        memory_dataset_id: str | None = None,
        learning_dataset_id: str| None = None,
    ):
        super().__init__(
            name=name,
            model_client=model_client,
            tools=tools,
            workbench=workbench,
            handoffs=handoffs,
            model_context=model_context,
            description=description,
            system_message=system_message,
            model_client_stream=model_client_stream,
            reflect_on_tool_use=reflect_on_tool_use,
            tool_call_summary_format=tool_call_summary_format,
            tool_call_summary_prompt=tool_call_summary_prompt,
            output_content_type=output_content_type,
            output_content_type_format=output_content_type_format,
            memory=memory,
            metadata=metadata,
            memory_function=memory_function,
            reply_function=reply_function,
            db_manager=db_manager,
            thread_id=thread_id,
            user_id=user_id,
            set_model_client=set_model_client,
            llm_mode_config=llm_mode_config,
        )

        self._developer_system_message = system_message

        # === workspace for assistant ===
        if not work_dir:
            if self._db_manager:
                DEFAULT_RUN_DIR: Path = self._db_manager.schema_manager.base_dir / "runs"
                self._work_dir = DEFAULT_RUN_DIR / self._user_id
            else:
                self._work_dir = Path(RUNS_DIR) / self._user_id
        else:
            self._work_dir = Path(work_dir) / self._user_id
        if not self._work_dir.exists():
            self._work_dir.mkdir(parents=True)

        # === initial UserProfileManager ===
        self._user_profile_manager = UserProfileManager(
            agent_name=self._name,
            work_dir=self._work_dir,
            user_id=self._user_id,
            thread_id=self._thread_id,
        )
        self._update_user_config_tools = [self._user_profile_manager.get_user_config_tool()]
        
        # === basic tools ===
        self._only_in_workspace = only_in_workspace
        self._extra_work_dirs = extra_work_dirs
        self._basic_funcs: List[Callable] = get_operator_funcs(
            work_dir, 
            only_in_workspace=self._only_in_workspace,
            extra_dirs = self._extra_work_dirs,
            )
        self._basic_funcs_names = [func.__name__ for func in self._basic_funcs]
        for func in self._basic_funcs:
            self._tools.append(FunctionTool(func, description=func.__doc__))

        # === model context ===
        self._token_limit = token_limit
        self._rag_flow_url = rag_flow_url
        self._rag_flow_token = rag_flow_token
        self._memory_dataset_id = memory_dataset_id
        self._learning_dataset_id = learning_dataset_id or memory_dataset_id
        self._memory_document_id = self._user_profile_manager.get_document_ids(self._thread_id)
        self._learning_document_id = self._user_profile_manager.get_document_ids(self._user_id)
        # memory manager
        self._model_context = DrSaiChatCompletionContext(
            agent_name=self._user_profile_manager.agent_name,
            model_client=self._model_client,
            user_id=self._user_id,
            thread_id=self._thread_id,
            work_dir=self._work_dir,
            token_limit=self._token_limit,
            rag_flow_url=self._rag_flow_url,
            rag_flow_token=self._rag_flow_token,
            dataset_id=self._memory_dataset_id,
            document_id=self._memory_document_id,
            learning_dataset_id=self._learning_dataset_id,
            learning_document_id=self._learning_document_id,
        )
        if not self._model_context._rag_flow_manager:
            raise ValueError("RAGFlowManager is not initialized in DrSaiChatCompletionContext")
        funcs = [self._model_context.retrieve_from_memory, self._model_context.summry_conversation_to_memory]
        for func in funcs:
            self._tools.append(FunctionTool(func, description=func.__doc__))
                
        # === skills ===
        self._skills_dir = skills_dir
        if self._user_profile_manager.first_time_setup and self._skills_dir:
            src_dirs = self._skills_dir if isinstance(self._skills_dir, list) else [self._skills_dir]
            dst_root = self._user_profile_manager.skills_dir
            for src_dir in src_dirs:
                src_path = Path(src_dir)
                for skill_folder in src_path.iterdir():
                    if skill_folder.is_dir():
                        dst = dst_root / skill_folder.name
                        if not dst.exists():
                            shutil.copytree(skill_folder, dst)
        self._agent_skills_tools = []

        # === executor ===
        self._local_executor = executor
        
        # === sub_agent_config ===
        self._sub_agent_config = sub_agent_config
        self._user_sub_agents = {}
        self._user_sub_agents.update(sub_agent_config)
        self._subagent_tools = []

        # === todo manager ===
        self._todo_manager = TodoManager()
        self._todo_tools = [get_todo_manager_tool()]

        # 初始化实例变量供edge_agent_core使用
        # self._current_plan = None
        # self._task_planner = None
        # self._memory_manager = None

        # max_turn_count
        self._max_turn_count = max_turn_count

        # config file mtime cache for lazy reloading
        self._config_mtimes: Dict[str, float] = {}
        self._cached_tools_prompt: str = ""
        self._cached_skills_loader = None

    def _file_changed(self, path: Path) -> bool:
        """Check if a file/dir has been modified since last check, updating the cached mtime."""
        try:
            mtime = path.stat().st_mtime
        except FileNotFoundError:
            return True
        key = str(path)
        if self._config_mtimes.get(key) != mtime:
            self._config_mtimes[key] = mtime
            return True
        return False

    def update_system_prompt(self, additional_prompt: str = "") -> str:
        """获取agent描述、用户画像并更新系统消息"""
        user_context = self._user_profile_manager.get_agent_system_prompt()
        enhanced_system_message = self._developer_system_message
        if user_context and self._developer_system_message:
            # 合并系统消息和用户上下文
            enhanced_system_message = f"""{self._developer_system_message}

{user_context}

Current Session_ID is {self._thread_id}
"""
        enhanced_system_message += additional_prompt
        self._system_messages = [SystemMessage(content=enhanced_system_message)]
    
    def update_user_skills(self) -> SkillLoader:
        """加载/更新用户技能"""
        skills_loader = None
        # 首先尝试从用户的skills目录加载
        user_skills_dir = self._user_profile_manager.skills_dir
        if user_skills_dir.exists() and list(user_skills_dir.glob("*/SKILL.md")):
            skills_loader = SkillLoader(skills_dir=str(user_skills_dir))
        # # 再从指定的skills目录加载
        # if self._skills_dir:
        #     if not skills_loader:
        #         skills_loader = SkillLoader(skills_dir=self._skills_dir)
        #     else:
        #         skills_loader.add_skills_by_dir(skills_dir=self._skills_dir)
        # 获取技能描述
        if skills_loader and skills_loader.skills:
            self._agent_skills_tools = [get_agent_skills_tool(descriptions=skills_loader.get_descriptions())]
        else:
            self._agent_skills_tools = []
        return skills_loader
    
    async def update_user_tools(self) -> str:
        """将用户的自定义配置工具接入到agent中"""
        user_mcp_tools = []
        user_local_tools = []
        tools_config = self._user_profile_manager.load_user_tools_config()
        for tool in tools_config:
            try:
                if tool.get("type") == "mcp-std":
                    config = tool.get("config")
                    if "command" in config and "args" in config:
                        std_mcp_tools = await mcp_server_tools(StdioServerParams(
                            command=config["command"],
                            args=config["args"]
                        ))
                        user_mcp_tools.extend(std_mcp_tools)
                elif tool.get("type") == "mcp-sse":
                    config = tool.get("config")
                    if "url" in config:
                        sse_mcp_tools = await mcp_server_tools(SseServerParams(
                            url=config["url"],
                            headers = config.get("headers", None),
                            timeout=config.get("timeout", float(20)),
                            sse_read_timeout=config.get("sse_read_timeout", float(300)),
                        ))
                        user_mcp_tools.extend(sse_mcp_tools)
                else:
                    user_local_tools.append(str(config)+"\n")

            except Exception as e:
                # print(f"Error loading tool: {e}")
                pass

        self._workbench._tools = self._tools + user_mcp_tools
        self._tools_names = [tool.name for tool in self._workbench._tools ]

        if user_local_tools:
            user_local_tools_prompt = "The info about the user's local function is as follows. When needed, you can execute it on the command line using `run_bash` tool\n\n"
            user_local_tools_prompt += "\n".join(user_local_tools)
        else:
            user_local_tools_prompt = ""
        return user_local_tools_prompt
    
    def get_subagent_descriptions(self, sub_agent_config: dict) -> str:
        """Generate agent type descriptions for system prompt."""
        return "\n".join(
            f"- {name}: {cfg['description']}"
            for name, cfg in sub_agent_config.items()
        )
    
    def update_user_subagents(self):
        """Update user subagents."""
        subagents_config = self._user_profile_manager.load_subagents_config()
        self._user_sub_agents.update(subagents_config)

        if self._user_sub_agents:
            self._sub_agent_descriptions = self.get_subagent_descriptions(sub_agent_config = self._user_sub_agents)
            self._subagent_tools = [
                get_subagent_tools(
                    sub_agents=list(self._user_sub_agents.keys()),
                    description=self._sub_agent_descriptions)]
        else:
            self._sub_agent_descriptions = ""
            self._subagent_tools = []

    async def run_stream(
        self,
        *,
        task: str | BaseChatMessage | Sequence[BaseChatMessage] | None = None,
        cancellation_token: CancellationToken | None = None,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | TaskResult, None]:
        """Run the agent with the given task and return a stream of messages
        and the final task result as the last item in the stream."""
        if cancellation_token is None:
            cancellation_token = CancellationToken()
        self._cancellation_token = cancellation_token
        input_messages: List[BaseChatMessage] = []
        output_messages: List[BaseAgentEvent | BaseChatMessage] = []
        if task is None:
            pass
        elif isinstance(task, str):
            text_msg = TextMessage(content=task, source="user", metadata={"internal": "yes"})
            # text_msg = TextMessage(content=task, source="user")
            input_messages.append(text_msg)
            output_messages.append(text_msg)
            yield text_msg
        elif isinstance(task, BaseChatMessage):
            task.metadata["internal"] = "yes"
            input_messages.append(task)
            output_messages.append(task)  
            yield task
        else:
            if not task:
                raise ValueError("Task list cannot be empty.")
            for msg in task:
                if isinstance(msg, BaseChatMessage):
                    msg.metadata["internal"] = "yes"
                    input_messages.append(msg)
                    output_messages.append(msg)
                    try:
                        attached_files_json = msg.metadata.get("attached_files")
                        if attached_files_json:
                            attached_files = json.loads(attached_files_json)
                            for file in attached_files:
                                download_file_from_url_or_base64(
                                    file_info = file, 
                                    save_path = f"{self._user_profile_manager.download_dir}/{file['name']}")
                        settings_config = msg.metadata.get("settings_config")
                        if settings_config:
                            settings_config = json.loads(settings_config)
                            default_config_name = settings_config.get("defult_config_name")
                            llm_name = self._llm_mode_config.get(default_config_name)
                            if llm_name != self._model_client._create_args["model"] and self._set_model_client:
                                self._model_client = self._set_model_client(default_config_name)
                    except Exception as e:
                        logger.error(f"Error processing message metadata: {e}")
                    yield msg
                else:
                    raise ValueError(f"Invalid message type in sequence: {type(msg)}")
        async for message in self.on_messages_stream(input_messages, cancellation_token):
            if isinstance(message, Response):
                yield message.chat_message
                output_messages.append(message.chat_message)
                yield TaskResult(messages=output_messages)
            else:
                yield message
                if isinstance(message, ModelClientStreamingChunkEvent):
                    # Skip the model client streaming chunk events.
                    continue
                output_messages.append(message)
                
    async def on_messages_stream(
        self, messages: Sequence[BaseChatMessage], cancellation_token: CancellationToken
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """
        Process the incoming messages with the assistant agent and yield events/responses as they happen.
        """

        # monitor the pause event
        if self.is_paused:
            yield Response(
                chat_message=TextMessage(
                    content=f"The {self.name} is paused.",
                    source=self.name,
                    metadata={"internal": "yes"},
                )
            )
            return

        inner_messages: List[BaseAgentEvent | BaseChatMessage] = []
        try:

            # initialize the learning memory document
            if self._user_profile_manager.first_time_setup:
                self._learning_document_id = await self._model_context.create_new_session_document(dataset_id = self._learning_dataset_id, create_type="learning_memory" )
                self._user_profile_manager.update_document_ids(thread_id=self._user_id, document_id=self._learning_document_id)
            # create the new session document
            if self._memory_document_id is None:
                self._memory_document_id = await self._model_context.create_new_session_document(
                    user_id = self._user_id,
                    thread_id = self._thread_id,
                    work_dir = self._work_dir,
                )
                self._user_profile_manager.update_document_ids(thread_id=self._thread_id, document_id=self._memory_document_id)
                self._model_context._document_id = self._memory_document_id


            # load/update tools only if TOOLS_CONFIG.json changed
            tools_changed = self._file_changed(self._user_profile_manager.tools_config_path)
            if tools_changed:
                self._cached_tools_prompt = await self.update_user_tools()

            # update system prompt if AGENTS.md or tools prompt changed
            if tools_changed or self._file_changed(self._user_profile_manager.agents_md):
                self.update_system_prompt(additional_prompt=self._cached_tools_prompt)

            # load/update skills only if skills directories changed
            skills_changed = self._file_changed(self._user_profile_manager.skills_dir)
            if self._skills_dir:
                extra_dirs = self._skills_dir if isinstance(self._skills_dir, list) else [self._skills_dir]
                skills_changed = skills_changed or any(self._file_changed(Path(d)) for d in extra_dirs)
            if skills_changed or self._cached_skills_loader is None:
                self._cached_skills_loader = self.update_user_skills()
            skills_loader = self._cached_skills_loader

            # load/update subagents only if SUBAGENT_CONFIG.json changed
            if self._file_changed(self._user_profile_manager.subagent_config_path):
                self.update_user_subagents()

            # manager ToolSchema
            manager_tools = self._update_user_config_tools+self._agent_skills_tools+self._subagent_tools+self._todo_tools

            # count the number of tools
            if isinstance(self._model_context, DrSaiChatCompletionContext):
                self._model_context._tool_schema = await self._workbench.list_tools()
                self._model_context._tool_schema += manager_tools

            # Gather all relevant state here
            agent_name = self._user_profile_manager.agent_name
            model_context = self._model_context
            memory = self._memory
            system_messages = self._system_messages
            workbench = self._workbench
            handoff_tools = self._handoff_tools
            handoffs = self._handoffs
            model_client = self._model_client
            model_client_stream = self._model_client_stream
            reflect_on_tool_use = self._reflect_on_tool_use
            tool_call_summary_format = self._tool_call_summary_format
            output_content_type = self._output_content_type
            format_string = self._output_content_type_format
            
            # Add new user/handoff messages to the model context
            await self._add_messages_to_context(
                model_context=model_context,
                messages=messages,
            )

            # TODO: Update model context with any relevant memory -> When? How?

            turn_count = 0
            while turn_count < self._max_turn_count:
                
                model_result = None
                async for inference_output in self._call_llm(
                    model_client=model_client,
                    model_client_stream=model_client_stream,
                    system_messages=system_messages,
                    model_context=model_context,
                    workbench=workbench,
                    handoff_tools=handoff_tools,
                    manager_tools=manager_tools,
                    agent_name=agent_name,
                    cancellation_token=cancellation_token,
                    output_content_type=output_content_type,
                ):
                    if self.is_paused:
                        raise asyncio.CancelledError()
                    
                    if isinstance(inference_output, CreateResult):
                        model_result = inference_output
                    else:
                        # Streaming chunk event
                        yield inference_output

                assert model_result is not None, "No model result was produced."

                # --- NEW: If the model produced a hidden "thought," yield it as an event ---
                if model_result.thought:
                    thought_event = ThoughtEvent(content=model_result.thought, source=agent_name)
                    yield thought_event
                    inner_messages.append(thought_event)

                # Add the assistant message to the model context (including thought if present)  
                await model_context.add_message(
                    AssistantMessage(
                        content=model_result.content,
                        source=self._name,
                        thought=getattr(model_result, "thought", None),
                    )
                )
                
                # If direct text response (string)
                if isinstance(model_result.content, str):
                    reponse = await self.handle_str_reponse(
                            model_result = model_result,
                            agent_name = agent_name,
                            format_string = format_string,
                            inner_messages = inner_messages,
                            output_content_type = output_content_type,)
                    if self._user_profile_manager.first_time_setup:
                        yield TextMessage(
                            content="\n\n(●'◡'●)如果您需要调整我的名称、我对您的称呼、您涉及领域，请告诉我，我来调整(If you need to adjust my name, how I address you, or your field of expertise, please let me know, and I will make the changes).",
                            source=agent_name,
                            metadata={"internal": "no"},
                        )
                        self._user_profile_manager.first_time_setup = False
                    yield reponse
                    break

                # Otherwise, we have function calls
                assert isinstance(model_result.content, list) and all(
                    isinstance(item, FunctionCall) for item in model_result.content
                )
                
                # handle tool call
                for i in range(len(model_result.content)):
                    argument = json.loads(model_result.content[i].arguments)
                    tool_name = model_result.content[i].name
                    call_id = model_result.content[i].id
                    if tool_name == "TodoWrite":
                        async for message in self.handle_todo_write(
                            argument = argument,
                            tool_name = tool_name,
                            call_id = call_id,
                            agent_name = agent_name, 
                            model_context = model_context):
                            if isinstance(message, StopMessage):
                                yield Response(
                                    chat_message=message,
                                    inner_messages=inner_messages,
                                )
                                return
                            yield message
                    elif tool_name == "Task":
                        async for message in self.handle_subagent_repsonse(
                            agent_name = agent_name,
                            model_client = model_client,
                            model_client_stream = model_client_stream,
                            model_context = model_context,
                            argument = argument,
                            tool_name = tool_name,
                            call_id = call_id,
                            cancellation_token = cancellation_token,
                            output_content_type = output_content_type,
                        ):
                            if isinstance(message, StopMessage):
                                yield Response(
                                    chat_message=message,
                                    inner_messages=inner_messages,
                                )
                                return
                            yield message
                    elif tool_name == "Skill":
                        skill_content = skills_loader.run_skill(argument["skill"])
                        # await model_context.add_message(
                        #     UserMessage(
                        #         content=f"Skill for {argument["skill"]}: {skill_content}",
                        #         source="user",
                        #     )
                        # )
                        await model_context.add_message(FunctionExecutionResultMessage(
                            content=[FunctionExecutionResult(
                                content = f"Skill for {argument["skill"]}:\n\n {skill_content}",
                                name = tool_name,
                                call_id = call_id,
                                is_error = False,
                            ),]
                        ))
                        yield AgentLogEvent(
                            title=f"I am reading skill: {argument["skill"]}.",
                            source=agent_name, 
                            content=str(argument), 
                            content_type="tools")
                        yield ToolCallSummaryMessage(
                            content=f"<think>Skill for {argument["skill"]}:\n\n {skill_content}</think>\n",
                            source=agent_name,
                        )
                    elif tool_name in self._tools_names:
                        async for message in self._process_model_result(
                            model_result=model_result,
                            inner_messages=inner_messages,
                            cancellation_token=cancellation_token,
                            agent_name=agent_name,
                            system_messages=system_messages,
                            model_context=model_context,
                            workbench=workbench,
                            handoff_tools=handoff_tools,
                            handoffs=handoffs,
                            model_client=model_client,
                            model_client_stream=model_client_stream,
                            reflect_on_tool_use=reflect_on_tool_use,
                            tool_call_summary_format=tool_call_summary_format,
                            tool_call_summary_prompt=self._tool_call_summary_prompt,
                            output_content_type=output_content_type,
                            format_string=format_string,
                        ):
                            if self.is_paused:
                                raise asyncio.CancelledError()
                            if isinstance(message, Response):
                                # yield ModelClientStreamingChunkEvent(content="<think>", source=agent_name)
                                # yield ModelClientStreamingChunkEvent(content=str(message.chat_message.content), source=agent_name)
                                # yield ModelClientStreamingChunkEvent(content="</think>\n", source=agent_name)
                                yield message.chat_message
                            else:
                                yield message
                    elif tool_name == "UpdateUserConfig":
                        update_message = self._user_profile_manager.update_user_config(**argument)
                        # await model_context.add_message(
                        #     UserMessage(
                        #         content=update_message,
                        #         source="user",
                        #     )
                        # )
                        await model_context.add_message(FunctionExecutionResultMessage(
                            content=[FunctionExecutionResult(
                                content = update_message,
                                name = tool_name,
                                call_id = call_id,
                                is_error = False,
                            ),]
                        ))
                        agent_name = self._user_profile_manager.agent_name
                        yield AgentLogEvent(
                            title=f"I am updating user's config.",
                            source=agent_name, 
                            content=str(argument), 
                            content_type="tools")
                    else:
                        await model_context.add_message(FunctionExecutionResultMessage(
                            content=[FunctionExecutionResult(
                                content = f"Unknown tool: {model_result.content[i].name}",
                                name = tool_name,
                                call_id = call_id,
                                is_error = False,
                            ),]
                        ))
                        # await model_context.add_message(
                        #     UserMessage(
                        #         content=f"Unknown tool: {model_result.content[i].name}",
                        #         source="user",
                        #     )
                        # )

                turn_count += 1
                if turn_count > self._max_turn_count:
                    yield Response(
                        chat_message=TextMessage(
                            content="\n\n(●'◡'●)抱歉，已达最大的任务循环次数，触发了保护措施，请重新调整您的询问方式或者更具体的告诉您的助手应该怎么做。",
                            source=agent_name,
                            metadata={"internal": "no"},
                        inner_messages=inner_messages,
                    ))
                    return

        except asyncio.CancelledError:
            # If the task is cancelled, we respond with a message.
            yield Response(
                chat_message=TextMessage(
                    content="The task was cancelled by the user.",
                    source=self._user_profile_manager.agent_name,
                    metadata={"internal": "yes"},
                ),
                inner_messages=inner_messages,
            )
        except Exception as e:
            logger.error(f"Error in {self._user_profile_manager.agent_name}: {e}")
            logger.error(traceback.format_exc())
            # add to chat history
            await self._model_context.add_message(
                AssistantMessage(
                    content=f"An error occurred while executing the task: {e}",
                    source=self._name
                )
            )
            yield Response(
                chat_message=TextMessage(
                    content=f"An error occurred while executing the task: {e}",
                    source=self._user_profile_manager.agent_name,
                    metadata={"internal": "no"},
                ),
                inner_messages=inner_messages,
            )
        finally:
            # save/update the conversation to {worker_dir}/memories
            if isinstance(self._model_context, DrSaiChatCompletionContext):
                if self._model_context._rag_flow_manager:
                    await self._model_context.upload_conversation_to_ragflow()
                self._model_context._current_messages = []
                current_session_memory = self._model_context._history_messages
                # TODO: use updates instead of full writes
                self._user_profile_manager.save_session_memory(current_session_memory)
                

    async def _call_llm(
        self,
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        system_messages: List[SystemMessage],
        model_context: ChatCompletionContext,
        workbench: Workbench,
        handoff_tools: List[BaseTool[Any, Any]],
        manager_tools: List[ToolSchema],
        agent_name: str,
        cancellation_token: CancellationToken,
        output_content_type: type[BaseModel] | None,
    ) -> AsyncGenerator[Union[CreateResult, ModelClientStreamingChunkEvent], None]:
        """
        Perform a model inference and yield either streaming chunk events or the final CreateResult.
        """

        if isinstance(model_context, DrSaiChatCompletionContext):
            all_messages = await model_context.get_messages(cancellation_token = cancellation_token)
        else:
            all_messages = await model_context.get_messages()
        
        llm_messages: List[LLMMessage] = self._get_compatible_context(model_client=model_client, messages=system_messages + all_messages)

        # 自定义的memory_function，用于RAG检索等功能，为大模型回复增加最新的知识
        if self._memory_function is not None:
            llm_messages = await self._call_memory_function(llm_messages, model_client, cancellation_token, agent_name)

        all_tools = (await workbench.list_tools()) + handoff_tools + manager_tools
        # model_result: Optional[CreateResult] = None
        if self._reply_function is not None:
            # 自定义的reply_function，用于自定义对话回复的定制
            async for chunk in self._call_reply_function(
                llm_messages, 
                model_client = model_client, 
                workbench=workbench,
                handoff_tools=handoff_tools,
                tools = all_tools,
                agent_name=agent_name, 
                cancellation_token=cancellation_token,
                db_manager=self._db_manager,
            ):
                # if isinstance(chunk, CreateResult):
                #     model_result = chunk
                yield chunk
        else:
           async for chunk in self.call_llm(
                agent_name = agent_name,
                model_client = model_client,
                llm_messages = llm_messages, 
                tools = all_tools, 
                model_client_stream = model_client_stream,
                cancellation_token = cancellation_token,
                output_content_type = output_content_type,
           ):
               yield chunk
    
    @classmethod
    async def _process_model_result(
        cls,
        model_result: CreateResult,
        inner_messages: List[BaseAgentEvent | BaseChatMessage],
        cancellation_token: CancellationToken,
        agent_name: str,
        system_messages: List[SystemMessage],
        model_context: ChatCompletionContext,
        workbench: Workbench,
        handoff_tools: List[BaseTool[Any, Any]],
        handoffs: Dict[str, HandoffBase],
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        reflect_on_tool_use: bool,
        tool_call_summary_format: str,
        tool_call_summary_prompt: str | None,
        output_content_type: type[BaseModel] | None,
        format_string: str | None = None,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage | Response, None]:
        """
        Handle final or partial responses from model_result, including tool calls, handoffs,
        and reflection if needed.
        """
        
        tool_call_msg = ToolCallRequestEvent(
            content=model_result.content,
            source=agent_name,
            models_usage=model_result.usage,
        )
        inner_messages.append(tool_call_msg)
        logger.debug(tool_call_msg)
        yield tool_call_msg
        tools_name = [tool.name for tool in model_result.content] 
        yield AgentLogEvent(
            title="I am using tools: " + " ".join(tools_name),
            source=agent_name, 
            content=str(tool_call_msg.content), 
            content_type="tools")

        # STEP 4B: Execute tool calls
        executed_calls_and_results = await asyncio.gather(
            *[
                cls._execute_tool_call(
                    tool_call=call,
                    workbench=workbench,
                    handoff_tools=handoff_tools,
                    agent_name=agent_name,
                    cancellation_token=cancellation_token,
                )
                for call in model_result.content
            ]
        )
        exec_results = [result for _, result in executed_calls_and_results]
        await model_context.add_message(FunctionExecutionResultMessage(content=exec_results))
        normal_tool_calls = [(call, result) for call, result in executed_calls_and_results if call.name not in handoffs]
        tool_call_summaries: List[str] = []
        for tool_call, tool_call_result in normal_tool_calls:
            tool_call_summaries.append(
                tool_call_summary_format.format(
                    tool_name=tool_call.name,
                    arguments=tool_call.arguments,
                    result=tool_call_result.content,
                )
            )
        tool_call_summary = "\n".join(tool_call_summaries)
        yield Response(
                chat_message=ToolCallSummaryMessage(
                    content="<think>The results of execution:\n " + tool_call_summary + "</think>\n",
                    source=agent_name,
                ),
                inner_messages=inner_messages,
            )

    async def handle_str_reponse(
            self,
            model_result: CreateResult,
            agent_name: str,
            format_string: str | None,
            inner_messages: List[BaseAgentEvent | BaseChatMessage],
            output_content_type: type[BaseModel] | None,
    ) -> Response:

        if output_content_type:
            content = output_content_type.model_validate_json(model_result.content)
            return Response(
                chat_message=StructuredMessage[output_content_type](  # type: ignore[valid-type]
                    content=content,
                    source=agent_name,
                    models_usage=model_result.usage,
                    format_string=format_string,
                ),
                inner_messages=inner_messages,
            )
        else:
            return Response(
                chat_message=TextMessage(
                    content=model_result.content,
                    source=agent_name,
                    models_usage=model_result.usage,
                ),
                inner_messages=inner_messages,
            )
    
    def get_tools_for_agent(self, agent_type: str) -> list:
        """Filter tools based on agent type."""
        allowed = self._user_sub_agents.get(agent_type, {}).get("tools", "*")
        if allowed == "*":
            return self._tools
        return [t for t in self._workbench._tools if t.name in allowed]

    async def handle_subagent_repsonse(
        self,
        agent_name: str,
        model_client: ChatCompletionClient,
        model_client_stream: bool,
        model_context: ChatCompletionContext,
        argument:Dict[str, Any],
        tool_name: str,
        call_id: str,
        cancellation_token: CancellationToken,
        output_content_type: type[BaseModel] | None,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage, None]:
        """
        Sub agent can actuate the sub task.

        The types of sub agent:
        1. code_executor
        2. normal drsai agent
        3. worker agent
        """
        try:
            description, prompt, sub_agent_name = argument["description"], argument["prompt"], argument["agent_type"]

            # get sub agent system prompt
            sub_system = f"""You are a {sub_agent_name} subagent at {self._work_dir}.

    {self._user_sub_agents[sub_agent_name].get("prompt", "")}

    Complete the task and return a clear, concise summary."""
            
            # construct task messages
            task_messages: Sequence[BaseChatMessage] = []
            llm_messages = await model_context.get_messages()
            # TODO: compress the background messages using  LLM
            backgroud_message = "Below are the historical chat records between the user and various intelligent assistants, which can be referenced when executing the current task.\n\n"
            for llm_message in llm_messages:
                if isinstance(llm_message, UserMessage) or isinstance(llm_message, AssistantMessage):
                    backgroud_message += f"{llm_message.source}: {llm_message.content}\n\n"
            task_messages.append(TextMessage(content=backgroud_message, source="user"))
            task_messages.append(TextMessage(content=f"Current task: \n\n{prompt}", source="user"))

            # Get agent
            if sub_agent_name in self._user_sub_agents:
                sub_agent = self._user_sub_agents[sub_agent_name]
                description=self._sub_agent_config[sub_agent_name].get("description", "")
                sub_agent_type = sub_agent.get("type")
                if sub_agent_type == "CodeExecutorAgent":
                    venv_path = sub_agent.get("venv_path")
                    if venv_path:
                        executor = create_local_venv(work_dir=venv_path)
                    else:
                        executor = self._local_executor or create_local_venv(work_dir=self._user_profile_manager.tmp_dir)
                    subagent = CodeExecutorAgent(
                        name=sub_agent_name,
                        code_executor=executor,
                        model_client_stream=model_client_stream,
                    )
                elif sub_agent_type == "DrSaiAgent":
                    tools = self.get_tools_for_agent(sub_agent_name)
                    subagent = DrSaiAgent(
                        name=sub_agent_name,
                        system_message=sub_system,
                        description=description,
                        tools=tools,
                        model_client=model_client,
                        model_client_stream=model_client_stream,
                        output_content_type=output_content_type,)
                elif sub_agent_type == "HepAIWorkerAgent":
                    model_remote_configs = sub_agent.get("model_remote_configs")
                    url = model_remote_configs.get("url", "https://aiapi.ihep.ac.cn/apiv2")
                    name = model_remote_configs.get("name")
                    subagent = HepAIWorkerAgent(
                        name=sub_agent_name,
                        description=description,
                        model_remote_configs={
                            "url": url,
                            "api_key": self._model_client._client.api_key,
                            "name": name
                        },
                        chat_id=self._thread_id,
                        run_info={"name": self._user_profile_manager.user_name, "email": self._user_id},

                    )

            
            # Process task
            #  TODO: handle turn count fro multi-turn task.
            # turn_count = 0
            # while turn_count < self._max_turn_count:
            #     turn_count += 1
            async for message in subagent.on_messages_stream(messages=task_messages, cancellation_token=cancellation_token):
                if isinstance(message, Response):
                    yield message.chat_message
                    # await model_context.add_message(
                    #     UserMessage(
                    #         content=str(message.chat_message.content),
                    #         source="user",
                    #     )
                    # )
                    await model_context.add_message(FunctionExecutionResultMessage(
                        content=[FunctionExecutionResult(
                            content = str(message.chat_message.content),
                            name = tool_name,
                            call_id = call_id,
                            is_error = False,
                        ),]
                    ))
                    return
                yield message

        except Exception as e:
            logger.exception(f"Error in {self.name}")
            yield ModelClientStreamingChunkEvent(
                content=str(e)+"\n\n",
                source=self.name,
            )
            # await model_context.add_message(
            #     UserMessage(
            #         content=str(e),
            #         source="user",
            #     )
            # )
            await model_context.add_message(FunctionExecutionResultMessage(
                content=[FunctionExecutionResult(
                    content = str(e),
                    name = tool_name,
                    call_id = call_id,
                    is_error = True,
                ),]
            ))
            yield StopMessage(
                content=str(e),
                source=agent_name,
            )

    async def handle_todo_write(
        self,
        argument: Dict[str, Any],
        tool_name: str,
        call_id: str,
        agent_name: str,
        model_context: ChatCompletionContext,
    ) -> AsyncGenerator[BaseAgentEvent | BaseChatMessage, None]:
        try:
            todo_list = self._todo_manager.update(argument["items"])
            # send stream message
            # yield ModelClientStreamingChunkEvent(
            #     content=todo_list+"\n\n",
            #     source=self._user_profile_manager.agent_name,
            # )
            # add message to model_context with user source
            await model_context.add_message(FunctionExecutionResultMessage(
                content=[FunctionExecutionResult(
                    content = self._todo_manager.get_task_prompt(),
                    name = tool_name,
                    call_id = call_id,
                    is_error = False,
                ),]
            ))
            # await model_context.add_message(
            #     UserMessage(
            #         content=self._todo_manager.get_task_prompt(),
            #         source="user",
            #     )
            # )
            # send text message to save to db in drsai ui
            yield TextMessage(
                content=todo_list,
                source=self._user_profile_manager.agent_name,
                metadata={"interal": "no"},
            )
        except Exception as e:
            logger.exception(f"Error in {self.name}")
            yield TextMessage(
                content=str(e)+"\n\n",
                source=self._user_profile_manager.agent_name,
                metadata={"interal": "no"},
            )
            # await model_context.add_message(
            #     UserMessage(
            #         content=str(e),
            #         source="user",
            #     )
            # )
            await model_context.add_message(FunctionExecutionResultMessage(
                content=[FunctionExecutionResult(
                    content = str(e),
                    name = tool_name,
                    call_id = call_id,
                    is_error = True,
                ),]
            ))
            yield StopMessage(
                content=str(e),
                source=agent_name,
            )

    # TODO: handle user tools and evenroment

    # TODO: handle skills and self-learning -> # TODO: 支持动态获取skills

    # TODO: fixed the config
    def _to_config(self) -> DrSaiAssistantConfig:
        """Convert the assistant agent to a declarative config."""

        return DrSaiAssistantConfig(
            name=self.name,
            model_client=self._model_client.dump_component(),
            tools=None,  # versionchanged:: v0.5.5  Now tools are not serialized, Cause they are part of the workbench.
            workbench=self._workbench.dump_component() if self._workbench else None,
            handoffs=list(self._handoffs.values()) if self._handoffs else None,
            model_context=self._model_context.dump_component(),
            memory=[memory.dump_component() for memory in self._memory] if self._memory else None,
            description=self.description,
            system_message=self._system_messages[0].content
            if self._system_messages and isinstance(self._system_messages[0].content, str)
            else None,
            model_client_stream=self._model_client_stream,
            reflect_on_tool_use=self._reflect_on_tool_use,
            tool_call_summary_format=self._tool_call_summary_format,
            tool_call_summary_prompt=self._tool_call_summary_prompt,
            structured_message_factory=self._structured_message_factory.dump_component()
            if self._structured_message_factory
            else None,
            metadata=self._metadata,
            # drsaiAgent specific
            db_manager_config=self._db_manager.dump_component(),
            thread_id=self._thread_id,
            user_id=self._user_id,
            # skills and executor
            skills_dir=self._skills_dir,
            work_dir=self._work_dir,
            only_in_workspace=self._only_in_workspace,
            extra_work_dirs=self._extra_work_dirs,
            executor=self._local_executor.dump_component(),
            sub_agent_config=self._sub_agent_config,
            max_turn_count=self._max_turn_count,
            token_limit=self._token_limit,
            rag_flow_url=self._rag_flow_url,
            rag_flow_token=self._rag_flow_token,
            memory_dataset_id=self._memory_dataset_id,
            learning_dataset_id=self._learning_dataset_id,
        )
    
    @classmethod
    def _from_config(
        cls, config: DrSaiAssistantConfig, 
        db_manager: DatabaseManager,
        memory_function: Callable = None,
        reply_function: Callable = None,
        **kwargs,
        ) -> Self:
        """Create an assistant agent from a declarative config."""
        if config.structured_message_factory:
            structured_message_factory = StructuredMessageFactory.load_component(config.structured_message_factory)
            format_string = structured_message_factory.format_string
            output_content_type = structured_message_factory.ContentModel

        else:
            format_string = None
            output_content_type = None

        return cls(
            name=config.name,
            model_client=ChatCompletionClient.load_component(config.model_client),
            workbench=Workbench.load_component(config.workbench) if config.workbench else None,
            handoffs=config.handoffs,
            model_context=ChatCompletionContext.load_component(config.model_context) if config.model_context else None,
            tools=[BaseTool.load_component(tool) for tool in config.tools] if config.tools else None,
            memory=[Memory.load_component(memory) for memory in config.memory] if config.memory else None,
            description=config.description,
            system_message=config.system_message,
            model_client_stream=config.model_client_stream,
            reflect_on_tool_use=config.reflect_on_tool_use,
            tool_call_summary_format=config.tool_call_summary_format,
            tool_call_summary_prompt=config.tool_call_summary_prompt,
            output_content_type=output_content_type,
            output_content_type_format=format_string,
            metadata=config.metadata,
            # drsaiAgent specific
            memory_function=memory_function,
            reply_function=reply_function,
            db_manager=db_manager,
            thread_id=config.thread_id,
            user_id=config.user_id,
            # skills and executor
            skills_dir=config.skills_dir,
            work_dir=config.work_dir,
            only_in_workspace=config.only_in_workspace,
            extra_work_dirs=config.extra_work_dirs,
            executor=CodeExecutor.load_component(config.executor),
            sub_agent_config = config.sub_agent_config,
            max_turn_count = config.max_turn_count,
            token_limit = config.token_limit,
            rag_flow_url = config.rag_flow_url,
            rag_flow_token = config.rag_flow_token,
            memory_dataset_id = config.memory_dataset_id,
            learning_dataset_id = config.learning_dataset_id,
            **kwargs,
        )



from abc import ABC, abstractmethod


from pydantic import BaseModel
from typing import Sequence, Any
from typing_extensions import Self
from autogen_core._component_config import Component, ComponentBase
from autogen_core import CancellationToken
from autogen_core.model_context import ChatCompletionContext
from autogen_agentchat.messages import BaseChatMessage

class BaseSensorConfig(BaseModel):
    name: str | None = None
    data_url: str | None = None
    api_key: str | None = None


class BaseSensor(ABC, ComponentBase[BaseModel]):

    component_type = "sensor"

    @abstractmethod
    async def get_data(
        self,
        serson_config: BaseSensorConfig,
        cancellation_token: CancellationToken | None = None,
        **kwargs: Any,
    ) -> Any:
        """
        Retrieve data from the sensor.

        Args:
            serson_config: The configuration for the sensor.
            cancellation_token: A cancellation token to cancel the operation.

        Returns:
            Any: The retrieved data.
        """
        ...

    @abstractmethod
    async def update_context(
        self,
        messages: Sequence[BaseChatMessage],
        model_context: ChatCompletionContext,
    ) -> Any:
        """
        Update the provided model context using retrieved data content.

        Args:
            model_context: The context to update.

        Returns:
            Any: The updated model context.
        """
        ...
    
    @abstractmethod
    async def clear(self) -> None:
        """Clear all entries from memory."""
        ...

    @abstractmethod
    async def close(self) -> None:
        """Clean up any resources used by the memory implementation."""
        ...



import requests
from typing import Any
import httpx
import os

from pydantic import BaseModel
from typing import List, Dict, Any
from typing_extensions import Self
from autogen_core._component_config import Component
from autogen_core import CancellationToken
from autogen_core.model_context import ChatCompletionContext
from autogen_core.models import SystemMessage, UserMessage, AssistantMessage
from autogen_core.memory._base_memory import (
    Memory, 
    MemoryContent, 
    MemoryQueryResult, 
    UpdateContextResult,
    MemoryMimeType
    )
import traceback
from loguru import logger


class RAGFlowMemoryConfig(BaseModel):
    """
    Configuration for ListMemory component.
    Args:
        name: The name of the memory instance (optional)
        RAGFLOW_URL: The URL of the RAGFlow API (default: "https://ragflow.ihep.ac.cn/")
        RAGFLOW_TOKEN: The API token for RAGFlow (default: "")
        dataset_ids: The IDs of the datasets to search (optional, but either this or document_ids must be set)
        document_ids: The IDs of the documents to search (optional, but either this or dataset_ids must be set)
        page: Page number (default: 1)
        page_size: Maximum number of chunks per page (default: 30)
        similarity_threshold: Minimum similarity score (default: 0.2)
        vector_similarity_weight: Weight of vector cosine similarity (default: 0.3)
        top_k: Number of chunks engaged in vector cosine computation (default: 1024)
        rerank_id: The ID of the rerank model (optional)
        keyword: Enable keyword-based matching (default: False)
        highlight: Enable highlighting of matched terms (default: False)
    """

    name: str | None = None
    RAGFLOW_URL: str = "https://ragflow.ihep.ac.cn"
    RAGFLOW_TOKEN: str = ""
    dataset_ids: list[str]|None = None
    document_ids: list[str]|None = None
    page: int = 1
    page_size: int = 30
    similarity_threshold: float = 0.2
    vector_similarity_weight: float = 0.3
    top_k: int = 1024
    rerank_id: int|None = None
    keyword: bool = False
    highlight: bool = False
    cross_languages: list[str] = []


class RAGFlowMemory(Memory, Component[RAGFlowMemoryConfig]):
    component_type = "memory"
    component_provider_override = "drsai.RAGFlowMemory"
    component_config_schema = RAGFlowMemoryConfig

    def __init__(
            self, 
            config: RAGFlowMemoryConfig,
            ) -> None:
        self._config = config
        self._name = config.name or "ragflow_memory"

    @property
    def name(self) -> str:
        """Get the memory instance identifier.

        Returns:
            str: Memory instance name
        """
        return self._name

    async def query(
        self,
        query: str | MemoryContent = "",
        cancellation_token: CancellationToken | None = None,
        config: RAGFlowMemoryConfig = None,
    ) -> MemoryQueryResult:
        """Return all memories without any filtering.

        Args:
            query: Ignored in this implementation
            cancellation_token: Optional token to cancel operation
            config: Optional configuration to use instead of the one provided at initialization

        Returns:
            MemoryQueryResult containing all stored memories
        """
        async def retrieve_chunks(
            question: str,
            RAGFLOW_URL: str,
            RAGFLOW_TOKEN: str,
            dataset_ids: list[str] = None,
            document_ids: list[str] = None,
            page: int = 1,
            page_size: int = 30,
            similarity_threshold: float = 0.2,
            vector_similarity_weight: float = 0.3,
            top_k: int = 1024,
            rerank_id: str = None,
            keyword: bool = False,
            highlight: bool = False,
            cross_languages: list[str] = [],
        ):
            """
            Retrieve chunks from specified datasets using RAGFlow retrieval API.

            Args:
                question: The user query or query keywords (required)
                RAGFLOW_URL: str,
                RAGFLOW_TOKEN: str,
                dataset_ids: The IDs of the datasets to search (optional, but either this or document_ids must be set)
                document_ids: The IDs of the documents to search (optional, but either this or dataset_ids must be set)
                page: Page number (default: 1)
                page_size: Maximum number of chunks per page (default: 30)
                similarity_threshold: Minimum similarity score (default: 0.2)
                vector_similarity_weight: Weight of vector cosine similarity (default: 0.3)
                top_k: Number of chunks engaged in vector cosine computation (default: 1024)
                rerank_id: The ID of the rerank model (optional)
                keyword: Enable keyword-based matching (default: False)
                highlight: Enable highlighting of matched terms (default: False)

            Returns:
                JSON response containing the retrieved chunks

            Raises:
                ValueError: If neither dataset_ids nor document_ids is provided
            """
            # Validate that at least one of dataset_ids or document_ids is provided
            if not dataset_ids and not document_ids:
                raise ValueError("Either dataset_ids or document_ids must be provided")

            url = f"{RAGFLOW_URL}/api/v1/retrieval"

            # Build request body
            body = {
                "question": question,
                "page": page,
                "page_size": page_size,
                "similarity_threshold": similarity_threshold,
                "vector_similarity_weight": vector_similarity_weight,
                "top_k": top_k,
                "keyword": keyword,
                "highlight": highlight,
                "cross_languages": cross_languages
            }

            # Add optional parameters if provided
            if dataset_ids:
                body["dataset_ids"] = dataset_ids
            if document_ids:
                body["document_ids"] = document_ids
            if rerank_id:
                body["rerank_id"] = rerank_id

            # Set up headers with authorization token and content type
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {RAGFLOW_TOKEN}"
            }

            # Make the async HTTP POST request
            async with httpx.AsyncClient() as client:
                response = await client.post(url, json=body, headers=headers)
                response.raise_for_status()
                response = response.json()
                if "data" in response:
                    data = response["data"]
                else:
                    # Log server message if available to aid debugging
                    logger.warning(f'No data found in response: {response.get("message")}')
                    data = {"chunks": []}
                

            return data
        try:
            if not config:
                config = self._config
            data = await retrieve_chunks(
                question=query,
                RAGFLOW_URL=config.RAGFLOW_URL,
                RAGFLOW_TOKEN=config.RAGFLOW_TOKEN,
                dataset_ids=config.dataset_ids,
                document_ids=config.document_ids,
                page=config.page,
                page_size=config.page_size,
                similarity_threshold=config.similarity_threshold,
                vector_similarity_weight=config.vector_similarity_weight,
                top_k=config.top_k,
                rerank_id=config.rerank_id,
                keyword=config.keyword,
                highlight=config.highlight,
                cross_languages=config.cross_languages,
            )
            results = [MemoryContent(content=chunk["content"], mime_type =MemoryMimeType.TEXT, metadata=chunk) for chunk in data["chunks"]]
            return MemoryQueryResult(results=results)
        except Exception as e:
            print(f"Error retrieving chunks from RAGFlow: {str(e)}")
            return MemoryQueryResult(results=[])
    
    async def update_context(
        self,
        model_context: ChatCompletionContext,
    ) -> UpdateContextResult:
        """Update the model context by appending memory content.

        This method mutates the provided model_context by adding all memories as a UserMessage.

        Args:
            model_context: The context to update. Will be mutated if memories exist.

        Returns:
            UpdateContextResult containing the memories that were added to the context
        """

        messages = await model_context.get_messages()
        if not messages:
            return UpdateContextResult(memories=MemoryQueryResult(results=[]))

        # Extract query from last message
        last_message = messages[-1]
        query_text = last_message.content if isinstance(last_message.content, str) else str(last_message)

        # Query memory and get results
        query_results = await self.query(query_text)

        if query_results.results:
            # Format results for context
            memory_strings = [f"{i}. {str(memory.content)}" for i, memory in enumerate(query_results.results, 1)]
            memory_context = "\nRelevant memory content:\n" + "\n".join(memory_strings)

            # Add to context
            await model_context.add_message(UserMessage(content=memory_context, source="MemoryManager"))

        return UpdateContextResult(memories=query_results)

    async def add(self, content: MemoryContent, cancellation_token: CancellationToken | None = None) -> None:
        """
        Add a new content to memory.

        Args:
            content: The memory content to add
            cancellation_token: Optional token to cancel operation
        """
        pass

    async def clear(self) -> None:
        """Clear all entries from memory."""
        pass

    async def close(self) -> None:
        """Clean up any resources used by the memory implementation."""
        pass

    @classmethod
    def _from_config(cls, config: RAGFlowMemoryConfig) -> Self:
        return cls(config=config)

    def _to_config(self) -> RAGFlowMemoryConfig:
        return self._config
    

class RAGFlowMemoryManager:
    """
    Functions to interact with RAGFlow Memory API
    - List datasets 
    - list_documents
    - retrieve chunks by question
    """
    def __init__(
            self,
            rag_flow_url: str,
            rag_flow_token: str,
            timeout: httpx.Timeout = httpx.Timeout(connect=10.0, read=120.0, write=10.0, pool=5.0)
    ):
        self.base_url = rag_flow_url
        self.api_key = rag_flow_token
        self.headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json"
        }
        self.timeout = timeout

    async def list_datasets(self) -> list[dict[str, Any]]:
        try:
            async with httpx.AsyncClient(timeout=self.timeout) as client:
                response = await client.get(f"{self.base_url}/api/v1/datasets", headers=self.headers)
                return response.json()["data"]
        except:
            return []
    
    async def list_documents(self, dataset_id: str) -> list[dict[str, Any]]:
        try:
            async with httpx.AsyncClient(timeout=self.timeout) as client:
                response = await client.get(f"{self.base_url}/api/v1/datasets/{dataset_id}/documents", headers=self.headers)
                return response.json()["data"]["docs"]
        except:
            return []
    
    async def add_files_to_dataset(
            self,
            dataset_id: str,
            files_path: str|List[str],
            ) -> dict[str, Any]:
        """
        Add content to dataset.
        
        Args:
            dataset_id: The ID of the dataset to which the documents will be uploaded.
            files_path: Path to the file(s) to upload. Can be a single path (str) or list of paths (List[str]).
            
        Returns:
            Response from the API.
        """
        url = f"{self.base_url}/api/v1/datasets/{dataset_id}/documents"
        
        headers = {
            "Authorization": f"Bearer {self.api_key}"
        }
        
        # Convert single file path to list
        if isinstance(files_path, str):
            files_path = [files_path]
            
        # Prepare files for upload
        files = []
        opened_files = []
        try:
            # Open all files
            for file_path in files_path:
                abs_file_path = os.path.abspath(file_path)
                file_handle = open(abs_file_path, 'rb')
                opened_files.append(file_handle)
                files.append(('file', (os.path.basename(abs_file_path), file_handle, 'application/octet-stream')))
            
            # Upload all files in a single request
            async with httpx.AsyncClient(timeout=self.timeout) as client:
                response = await client.post(url, headers=headers, files=files)
                response.raise_for_status()
                return response.json()
        except Exception as e:
            print(f"Error uploading file(s) to dataset: {str(e)}")
            return {"error": str(e)}
        finally:
            # Close all opened files
            for file_handle in opened_files:
                try:
                    file_handle.close()
                except:
                    pass
    
    async def parse_files(self, dataset_id: str, document_ids: List[str]):
        """
        Parse documents in a specified dataset.
        
        Args:
            dataset_id: The dataset ID.
            document_ids: The IDs of the documents to parse.
            
        Returns:
            Response from the API.
        """
        url = f"{self.base_url}/api/v1/datasets/{dataset_id}/chunks"
        
        try:
            async with httpx.AsyncClient(timeout=self.timeout) as client:
                response = await client.post(
                    url, 
                    headers=self.headers, 
                    json={"document_ids": document_ids}
                )
                response.raise_for_status()
                return response.json()
        except Exception as e:
            print(f"Error parsing files: {str(e)}")
            return {"error": str(e)}
    
    async def update_document(
        self,
        dataset_id: str,
        document_id: str,
        name: str = None,
        meta_fields: Dict[str, Any] = None,
        chunk_method: str = None,
        parser_config: Dict[str, Any] = None
    ) -> dict[str, Any]:
        """
        Update configurations for a specified document.
        
        Args:
            dataset_id: (Path parameter)
                The ID of the associated dataset.
            document_id: (Path parameter)
                The ID of the document to update.
            "name": (Body parameter), string name of the document
            "meta_fields": (Body parameter), dict[str, Any] The meta fields of the document.
            "chunk_method": (Body parameter), string
                The parsing method to apply to the document:
                "naive": General
                "manual: Manual
                "qa": Q&A
                "table": Table
                "paper": Paper
                "book": Book
                "laws": Laws
                "presentation": Presentation
                "picture": Picture
                "one": One
                "email": Email
            "parser_config": (Body parameter), object
                The configuration settings for the dataset parser. The attributes in this JSON object vary with the selected "chunk_method":
                If "chunk_method" is "naive", the "parser_config" object contains the following attributes:
                    "chunk_token_num": Defaults to 256.
                    "layout_recognize": Defaults to true.
                    "html4excel": Indicates whether to convert Excel documents into HTML format. Defaults to false.
                    "delimiter": Defaults to "\n".
                    "task_page_size": Defaults to 12. For PDF only.
                    "raptor": RAPTOR-specific settings. Defaults to: {"use_raptor": false}.
                If "chunk_method" is "qa", "manuel", "paper", "book", "laws", or "presentation", the "parser_config" object contains the following attribute:
                    "raptor": RAPTOR-specific settings. Defaults to: {"use_raptor": false}.
                If "chunk_method" is "table", "picture", "one", or "email", "parser_config" is an empty JSON object.
            
        Returns:
            Response from the API.
        """
        url = f"{self.base_url}/api/v1/datasets/{dataset_id}/documents/{document_id}"
        
        # Prepare request body
        body = {}
        if name is not None:
            body["name"] = name
        if meta_fields is not None:
            body["meta_fields"] = meta_fields
        if chunk_method is not None:
            body["chunk_method"] = chunk_method
        if parser_config is not None:
            body["parser_config"] = parser_config
        
        try:
            async with httpx.AsyncClient(timeout=self.timeout) as client:
                response = await client.put(
                    url,
                    headers=self.headers,
                    json=body
                )
                response.raise_for_status()
                return response.json()
        except Exception as e:
            print(f"Error updating document: {str(e)}")
            return {"error": str(e)}

    async def add_files_to_dataset_and_parse(
            self,
            dataset_id: str,
            files_path: str|List[str],
            ) -> List[str]:
        """
        add files to dataset and parse
        Args:
            dataset_id: The ID of the dataset to which the documents will be uploaded.
            files_path: Path to the file(s) to upload. Can be a single path (str) or list of paths (List[str]).
            
        Returns:
            Response from the API.
        """
        result = await self.add_files_to_dataset(dataset_id, files_path)
        if result["code"] != 0:
            raise result["message"]
        document_ids = [datai["id"] for datai in result["data"]]
        result = await self.parse_files(dataset_id, document_ids)
        if result["code"] != 0:
            raise result["message"]
        return document_ids
    
    async def add_chunks_to_dataset(
            self,
            dataset_id: str,
            document_id: str,
            content: str,
            important_keywords: List[str] = None,
            questions: List[str] = None
            ) -> dict[str, Any]:
        """
        Adds a chunk to a specified document in a specified dataset.
        
        Args:
            dataset_id: The associated dataset ID.
            document_id: The associated document ID.
            content: The text content of the chunk.
            important_keywords: The key terms or phrases to tag with the chunk.
            questions: If there is a given question, the embedded chunks will be based on them.
            
        Returns:
            Response from the API.
        """
        url = f"{self.base_url}/api/v1/datasets/{dataset_id}/documents/{document_id}/chunks"
        
        # Prepare request body
        body = {
            "content": content
        }
        
        if important_keywords is not None:
            body["important_keywords"] = important_keywords
            
        if questions is not None:
            body["questions"] = questions
        
        try:
            async with httpx.AsyncClient(timeout=self.timeout) as client:
                response = await client.post(
                    url, 
                    headers=self.headers, 
                    json=body
                )
                response.raise_for_status()
                return response.json()
        except Exception as e:
            print(f"Error adding chunks to dataset: {str(e)}")
            return {"error": str(e)}
    
    async def retrieve_chunks_by_content(
            self,
            question: str,
            dataset_ids: list[str] = [],
            document_ids: list[str] = [],
            similarity_threshold: float = 0.2,
            vector_similarity_weight: float = 0.3,
            **kwargs: Any
            ) -> dict[str, Any]:
        """
        Retrieve chunks by question.
        kwargs:
            question: str
            dataset_ids: list[str]
            document_ids: list[str]
            page: int
            page_size: int
            similarity_threshold: float = 0.2
            vector_similarity_weight: float = 0.3
            top_k: int = 1024
            rerank_id: str = "hepai/bge-reranker-v2-m3___OpenAI-API@OpenAI-API-Compatible"
            keyword: bool
            highlight: bool
            cross_languages: list[str] = ["English", "Chinese"]
            metadata_condition: dict[str, Any]
        """
        params = {
            "question": question,
            "dataset_ids": dataset_ids,
            "document_ids": document_ids,
            "similarity_threshold": similarity_threshold,
            "vector_similarity_weight": vector_similarity_weight,
            **kwargs
        }
        try:
            if not dataset_ids and not document_ids:
                raise
            async with httpx.AsyncClient(timeout=self.timeout) as client:
                response = await client.post(
                    f"{self.base_url}/api/v1/retrieval", 
                    headers=self.headers,
                    json=params
                )
            return response.json()["data"]
        except Exception as e:
            logger.error("Error in retrieve_chunks:")
            logger.error(traceback.format_exc())
            return {}
if __name__ == "__main__":

    import json
    import asyncio

    base_url = "https://ragflow.ihep.ac.cn"
    api_key = "ragflow-*********" 
    ragflow_memory = RAGFlowMemoryManager(base_url, api_key)

    # list the datasets
    # datasets = asyncio.run(ragflow_memory.list_datasets())
    # print(json.dumps(datasets, indent=4, ensure_ascii=False))

    # list the documents
    # documents = asyncio.run(ragflow_memory.list_documents("70722df8519011f08a170242ac120006"))
    # print(json.dumps(documents, indent=4, ensure_ascii=False))

    # search content from datasets
    # result = asyncio.run(ragflow_memory.retrieve_chunks_by_content(
    #     question="北京出差的报销标准",
    #     dataset_ids=["70722df8519011f08a170242ac120006"]
    # ))
    # print(json.dumps(result, indent=4, ensure_ascii=False))

    # add files to dataset
    # file_path = "/home/xiongdb/drsai/README.md"
    # result = asyncio.run(ragflow_memory.add_files_to_dataset(
    #     dataset_id="70722df8519011f08a170242ac120006",
    #     files_path=file_path
    # ))
    # print(json.dumps(result, indent=4, ensure_ascii=False))

    # parse files
    # result = asyncio.run(ragflow_memory.parse_files(
    #     dataset_id="70722df8519011f08a170242ac120006",
    #     document_ids=["424b7bb8c8e711f091ce0242ac120006"]
    # ))
    # print(json.dumps(result, indent=4, ensure_ascii=False))
    
    # add chunks to dataset
    # dataset_id="70722df8519011f08a170242ac120006"
    # document_id="424b7bb8c8e711f091ce0242ac120006"
    # result = asyncio.run(ragflow_memory.add_chunks_to_dataset(
    #     dataset_id=dataset_id,
    #     document_id=document_id,
    #     content="opendrsai常用于专业科学智能体开发",
    #     important_keywords=["Opendrsai"],
    #     questions=["opendrsai常用于做什么"]
    # ))
    # print(json.dumps(result, indent=4, ensure_ascii=False))


from .base_sensor import BaseSensor, BaseSensorConfig
from drsai.modules.components.memory.ragflow_memory import RAGFlowMemoryManager
from typing import List, Dict,Sequence, Any
from typing_extensions import Self
from autogen_core._component_config import Component, ComponentBase
from autogen_core import CancellationToken
from autogen_core.model_context import ChatCompletionContext
from autogen_agentchat.messages import (
    BaseChatMessage,
    TextMessage,
    MultiModalMessage,
    )

class RAGFlowSensorConfig(BaseSensorConfig):
    rag_flow_url: str|None = None
    rag_flow_token: str|None = None
    dataset_id: str|None = None

class RAGFlowSensor(BaseSensor, Component[RAGFlowSensorConfig]):

    component_type = "sensor"
    component_provider_override = "drsai.RAGFlowSensor"
    component_config_schema = RAGFlowSensorConfig

    def __init__(
            self,
            rag_flow_url: str,
            rag_flow_token: str,
            dataset_id: str|None = None,
            ):
       
       self._rag_manager = RAGFlowMemoryManager(
            rag_flow_url=rag_flow_url,
            rag_flow_token=rag_flow_token
        )
       
       self._dataset_id = dataset_id
       
    async def upload_user_files(
            self, 
            messages: Sequence[BaseChatMessage],
            ):
        """
        messages: 
            为前端用户发送的消息列表，位于python/packages/drsai_ui/src/drsai_ui/ui_backend/backend/utils/utils.py，返回messages_return，包括
            [
            TextMessage(source="user", content=combined_text, metadata={"internal": "yes"}),
            MultiModalMessage(source="user",content=[query, *images],metadata={"attached_files": attached_files_json},),
            TextMessage(source="user",content=query,metadata={"attached_files": attached_files_json},)
            ]


        """

        for message in messages:
            if message.content.type == "file":
                await self._rag_manager.add_files_to_dataset_and_parse(
                    dataset_id=self._dataset_id,
                    files_path=message.content.file_path,
                )

    async def update_context(
        self,
        messages: Sequence[BaseChatMessage],
        model_context: ChatCompletionContext,
    ) -> Any:
        """
        Update the provided model context using retrieved data content.

        Args:
            messages: The messages to use for updating the model context.
            model_context: The context to update.

        Returns:
            Any: The updated model context.
        """
        pass
    
    async def get_data(
        self,
        serson_config: BaseSensorConfig,
        cancellation_token: CancellationToken | None = None,
        **kwargs: Any,
    ) -> Any:
        """
        Retrieve data from the sensor.

        Args:
            serson_config: The configuration for the sensor.
            cancellation_token: A cancellation token to cancel the operation.

        Returns:
            Any: The retrieved data.
        """
        pass

    async def clear(self) -> None:
       """Clear all entries from memory."""
       pass

    async def close(self) -> None:
        """Clean up any resources used by the memory implementation."""
        pass



"""
wechat_client.py — ilink Bot API 异步封装（httpx 版）
=====================================================
将 wechat_api.py 的同步 requests 实现移植为 httpx 异步版本。

三个核心接口：
  1. POST /ilink/bot/getupdates     —— 长轮询拉取新消息
  2. POST /ilink/bot/sendmessage    —— 向用户发送消息
"""

import asyncio
import base64
import json
import os
import secrets
import time
import uuid

import httpx

from drsai.configs.constant import WECHAT_DIR

BASE_URL = "https://ilinkai.weixin.qq.com"
SYNC_BUF_FILE = os.path.join(WECHAT_DIR, "sync_buf.txt")

MAX_LEN = 2048          # 微信单条消息建议不超过 2048 字符
MAX_DEDUP_SIZE = 1000   # 最多缓存最近 N 条消息 ID 用于去重


# ── 消息类型常量 ──────────────────────────────────────────────────────────────

class MessageType:
    USER = 1   # 用户发来的消息
    BOT  = 2   # Bot 发出的消息


class MessageItemType:
    TEXT  = 1
    IMAGE = 2
    VOICE = 3
    FILE  = 4
    VIDEO = 5


class MessageState:
    NEW        = 0
    GENERATING = 1
    FINISH     = 2


# ── 工具函数 ──────────────────────────────────────────────────────────────────

def _gen_uin() -> str:
    """生成随机 X-WECHAT-UIN（4字节随机数的 base64），用于请求头。"""
    return base64.b64encode(secrets.token_bytes(4)).decode()


def _gen_client_id() -> str:
    """生成唯一 client_id，防止重复发送。"""
    return f"py-bot-{int(time.time() * 1000)}-{uuid.uuid4().hex[:8]}"


# ── sync_buf 游标持久化 ───────────────────────────────────────────────────────

def load_sync_buf() -> str:
    """从本地文件加载消息游标（程序重启后可从上次中断处继续）。"""
    try:
        with open(SYNC_BUF_FILE, encoding="utf-8") as f:
            return f.read().strip()
    except FileNotFoundError:
        return ""


def save_sync_buf(buf: str) -> None:
    """持久化消息游标，防止重启后重复收到历史消息。"""
    os.makedirs(WECHAT_DIR, exist_ok=True)
    with open(SYNC_BUF_FILE, "w", encoding="utf-8") as f:
        f.write(buf)


# ── 文字分段 ─────────────────────────────────────────────────────────────────

def split_text(text: str, max_len: int = MAX_LEN) -> list[str]:
    """将超长文字按段落拆分，优先从换行处切割。"""
    if len(text) <= max_len:
        return [text]
    chunks = []
    while text:
        if len(text) <= max_len:
            chunks.append(text)
            break
        split_at = text.rfind("\n", 0, max_len)
        if split_at < max_len * 0.3:
            split_at = max_len
        chunks.append(text[:split_at])
        text = text[split_at:].lstrip("\n")
    return chunks


# ── 核心异步 API 类 ───────────────────────────────────────────────────────────

class AsyncWeChatAPI:
    """
    封装 ilink Bot 的三个核心接口（异步 httpx 版）。

    使用方式：
        api = AsyncWeChatAPI(bot_token="your_token")
        resp = await api.get_updates()
        await api.send_text(account_id, to_user, ctx_token, "你好")
    """

    def __init__(self, bot_token: str, base_url: str = BASE_URL):
        self.bot_token = bot_token
        self.base_url = base_url.rstrip("/")

    def _headers(self) -> dict:
        """构造每次请求必需的认证头。"""
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.bot_token}",
            "AuthorizationType": "ilink_bot_token",
            "X-WECHAT-UIN": _gen_uin(),
        }

    async def _post(self, path: str, body: dict, timeout: int = 40) -> dict:
        """通用异步 POST 请求，返回解析后的 JSON。"""
        url = f"{self.base_url}/{path}"
        async with httpx.AsyncClient(timeout=timeout) as client:
            resp = await client.post(url, json=body, headers=self._headers())
            resp.raise_for_status()
            return resp.json()

    # ── 接口1：拉取消息（长轮询） ─────────────────────────────────────────────

    async def get_updates(self, sync_buf: str | None = None) -> dict:
        """
        长轮询拉取新消息，服务端最多等待 ~30 秒后返回。

        参数：
            sync_buf: 上次返回的游标字符串（首次传 None）

        返回：
            {
              "ret": 0,
              "get_updates_buf": "...",   # 下次调用时传入
              "msgs": [...]               # 新消息列表，可能为空
            }
        """
        body = {"get_updates_buf": sync_buf} if sync_buf else {}
        return await self._post("ilink/bot/getupdates", body, timeout=40)

    # ── 接口2：发送文字消息 ───────────────────────────────────────────────────

    async def send_text(
        self,
        bot_account_id: str,
        to_user_id: str,
        context_token: str,
        text: str,
    ) -> dict:
        """
        发送文字消息给指定用户。

        参数：
            bot_account_id: 登录后得到的 account_id（ilink_bot_id）
            to_user_id:     目标用户的 ilink_user_id
            context_token:  从收到的消息中取出，用于关联会话
            text:           要发送的文字内容
        """
        body = {
            "msg": {
                "from_user_id": bot_account_id,
                "to_user_id": to_user_id,
                "client_id": _gen_client_id(),
                "message_type": MessageType.BOT,
                "message_state": MessageState.FINISH,
                "context_token": context_token,
                "item_list": [
                    {
                        "type": MessageItemType.TEXT,
                        "text_item": {"text": text},
                    }
                ],
            }
        }
        return await self._post("ilink/bot/sendmessage", body)

    # ── 辅助：从消息 item_list 中提取文本 ────────────────────────────────────

    @staticmethod
    def extract_text(msg: dict) -> str:
        """从 WeixinMessage 对象中提取所有文字内容，拼接返回。"""
        texts = []
        for item in msg.get("item_list") or []:
            if item.get("type") == MessageItemType.TEXT:
                texts.append(item.get("text_item", {}).get("text", ""))
        return "\n".join(filter(None, texts))

    @staticmethod
    def split_text(text: str, max_len: int = MAX_LEN) -> list[str]:
        """将超长文字按段落拆分。"""
        return split_text(text, max_len)



from autogen_agentchat.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    StructuredMessageFactory,
    MessageFactory,
    )
from autogen_core import FunctionCall
from autogen_core.models import (
    UserMessage, 
    SystemMessage, 
    AssistantMessage, 
    FunctionExecutionResultMessage,
    FunctionExecutionResult,
    )
from typing import (
    Any, 
    Dict, 
    Union, 
    Generic, 
    List, 
    Literal, 
    Mapping, 
    Optional, 
    Type, 
    TypeVar, 
    Sequence)
from enum import Enum 
from pydantic import BaseModel, Field, computed_field
import time
from drsai.modules.components.task_manager.base_task_system import TaskStatus

# Events

## Task
class TaskEvent(BaseAgentEvent):
    content: str|Dict[str, Any]
    type: Literal["TaskEvent"] = "TaskEvent"
    def to_text(self) -> str:
        if isinstance(self.content, str):
            return self.content
        else:
            return str(self.content)

## Files
class FileInfo(BaseModel):
    """A file info"""
    name: str
    url: Optional[str] = None
    base64_content: Optional[str] = None
    size: Optional[int] = None
    mime_type: Optional[str] = None
    description: Optional[str] = None
    download_method: Literal["url", "base64", "none"] = "url"

class FilesContent(BaseModel):
    """FilesEvent structure"""
    files: List[FileInfo]  # 文件列表
    title: Optional[str] = None  # 文件集合标题
    description: Optional[str] = None  # 集合描述
    send_time_stamp: float = Field(default_factory=time.time)

class FilesEvent(BaseAgentEvent):
    content: Union[str, FilesContent]
    type: Literal["FilesEvent"] = "FilesEvent"
    def to_text(self) -> str:
        if isinstance(self.content, str):
            return self.content
        else:
            return self.content.model_dump_json()

## Log
class Send_level(str, Enum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    DEBUG = "DEBUG"
    TRACE = "TRACE"
    FATAL = "FATAL"
    
class AgentLogEvent(BaseAgentEvent):
    """An event signaling a text output chunk from a model client in streaming mode."""
    title: str 
    content: str|List[FunctionCall|FunctionExecutionResult] = ""
    content_type: str|None = None
    """The type of content, such as web search results, etc."""
    send_time_stamp: float = Field(default_factory=time.time)
    send_level: Send_level = Send_level.INFO
    type: Literal["AgentLogEvent"] = "AgentLogEvent"

    def to_text(self) -> str:
        return f"[{time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(self.send_time_stamp))}] [{self.send_level.value}] {self.content}"

## Tool Long Task
class ToolLongTaskEvent(BaseAgentEvent):
    """An event signaling a text output chunk from a model client in streaming mode."""

    content: str|Dict[str, Any]
    tool_name: str|None = None
    task_status: TaskStatus|str = TaskStatus.queued.value
    """The name of the tool being used."""
    send_time_stamp: float = Field(default_factory=time.time)
    type: Literal["ToolLongTaskEvent"] = "ToolLongTaskEvent"
    def to_text(self) -> str:
        return f"[{time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(self.send_time_stamp))}] [{self.tool_name}] [{self.task_status}] {self.content}"


# Messages

class LongTaskQueryMessage(BaseChatMessage):
    """An Agent level message that can be used to send long task messages to the user"""

    content: str|Dict[str, Any]
    """The content of the message."""
    query_arguments: Dict[str, Any]|None = None
    """The query arguments for the long task."""
    tool_name: str|None = None
    """The name of the tool being used."""
    
    def to_text(self) -> str:
        if isinstance(self.content, str):
            return self.content
        else:
            return self.content.model_dump_json()

    def to_model_text(self) -> str:
        if isinstance(self.content, str):
            return self.content
        else:
            return self.content.model_dump_json()

    def to_model_message(self) -> UserMessage:
        content = self.to_text()
        return UserMessage(content=content, source=self.source)
    
    type: Literal["LongTaskQueryMessage"] = "LongTaskQueryMessage"

class AgentLongTaskMessage(BaseChatMessage):
    """An Agent level message that can be used to send long task messages to the user"""

    content: str|Dict[str, Any]
    """The content of the message."""
    task_status: TaskStatus|str = TaskStatus.queued.value
    """The status of the task."""
    query_arguments: Dict[str, Any]|None = None
    """The query arguments for the long task."""
    tool_name: str|None = None
    """The name of the tool being used."""
    
    def to_text(self) -> str:
        if isinstance(self.content, str):
            return self.content
        else:
            return self.content.model_dump_json()

    def to_model_text(self) -> str:
        if isinstance(self.content, str):
            return self.content
        else:
            return self.content.model_dump_json()

    def to_model_message(self) -> UserMessage:
        content = self.to_text()
        return UserMessage(content=content, source=self.source)
    
    type: Literal["AgentLongTaskMessage"] = "AgentLongTaskMessage"

# Message Factory

class DrSaiMessageFactory(MessageFactory):

    def __init__(self, custom_message_types: Sequence[BaseAgentEvent|BaseChatMessage]|None = None):
        super().__init__()

        self.register(TaskEvent)
        self.register(FilesEvent)
        self.register(AgentLogEvent)
        self.register(ToolLongTaskEvent)
        self.register(AgentLongTaskMessage)
        self.register(LongTaskQueryMessage)

        for message_type in custom_message_types or []:
            self.register(message_type)



"""
idle_monitor.py — 空闲 Agent 自动关闭监控
==========================================
后台任务：每隔 check_interval 秒扫描一次会话状态，
将超过 idle_seconds 无活动的 agent 实例关闭，以释放资源。

关闭后 agent 并不会被删除（session 记录仍在），
下次用户发消息时会通过 lazy_init 自动重建。
"""

import asyncio
import logging
from typing import TYPE_CHECKING

from .session_manager import SessionManager

if TYPE_CHECKING:
    from drsai.backend.run import DrSaiWorkerModel

logger = logging.getLogger(__name__)


async def idle_monitor(
    model: "DrSaiWorkerModel",
    session_manager: SessionManager,
    idle_seconds: int = 60*10,
    check_interval: int = 30,
) -> None:
    """
    持续运行的后台协程。

    参数：
        model          : DrSaiWorkerModel 实例（持有 drsai.agent_instance 字典）
        session_manager: 会话管理器（提供 get_idle 查询）
        idle_seconds   : 超过此秒数无活动则关闭 agent（默认 60s）
        check_interval : 每隔此秒数检查一次（默认 30s）
    """
    logger.info(
        "idle_monitor 已启动：idle_seconds=%ds, check_interval=%ds",
        idle_seconds,
        check_interval,
    )

    while True:
        try:
            await asyncio.sleep(check_interval)
        except asyncio.CancelledError:
            logger.info("idle_monitor 收到取消信号，退出。")
            break

        idle_ids = session_manager.get_idle(idle_seconds)
        if not idle_ids:
            continue

        for chat_id in idle_ids:
            if chat_id not in model.drsai.agent_instance:
                continue
            try:
                result = await model.close(chat_id)
                if result.get("status"):
                    logger.info("已关闭空闲 agent: %s（超过 %ds 无活动）", chat_id, idle_seconds)
                else:
                    logger.warning("关闭 agent %s 失败: %s", chat_id, result.get("message"))
            except Exception as e:
                logger.exception("关闭 agent %s 时出错: %s", chat_id, e)
